/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <getopt.h>

#include "common.h"
#include "options.h"
#include "log.h"
#include "util.h"
#include "core.h"

#define DEFAULT_OWNER_STATEFILENAME  "adeona-retrievecredentials"
#define DEFAULT_CLIENT_STATEFILENAME "adeona-clientstate"
#define OST_PWDSTR_BUFLEN 128

static inline void usage( char* prog )
{
    printf( "Usage: %s [options]\n"
            "Options:\n"
            "-p password   \t" "Specify password to use to encrypt owner statefile.\n"
			"-l logdir     \t" "Directory where output logs and data are stored\n"
            "-r resourcedir\t" "Directory containing resources (e.g. options.adeona)\n"
            "-s statefile\t"   "Initialize client and owner state; generating statefile.cst and statefile.ost\n"
            "\n", prog );
}

static void do_init( char* statefilename IN, char* pwdstr IN, int plen IN )
{   
    char* cstfilename = NULL;
    char* ostfilename = NULL;

    if( statefilename != NULL ) {
        cstfilename = concatstr( statefilename, ".cst" );
        ostfilename = concatstr( statefilename, ".ost" );
    }
    else {
        cstfilename = concatstr( DEFAULT_CLIENT_STATEFILENAME, ".cst" );
        ostfilename = concatstr( DEFAULT_OWNER_STATEFILENAME, ".ost" );
    }
        
    state_t* state = state_alloc();
        
    if( state_makenew( state ) == FAILURE ) {
        state_free( state );
        free( cstfilename );
        free( ostfilename );
        fatal( "could not generate an initial state" );
    }
    
    if( state_save_filename( cstfilename, state, NULL, 0, SC_ALL ) == FAILURE ) {
        state_free( state );
        free( cstfilename );
        free( ostfilename );
        fatal( "could not save client state file" );
    }
    
    FILE* ostfile;
    if( ( ostfile = fopen( ostfilename, "w" ) ) == NULL ) {
        state_free( state );
        free( cstfilename );
        free( ostfilename );
        fatal( "could not save owner state file" );
    }

    if( (write_options_file( ostfile, 1 ) == FAILURE) || (state_save( ostfile, state, pwdstr, plen, SC_SAVED ) == FAILURE) )  {
        state_free( state );
        free( cstfilename );
        free( ostfilename );
        fclose( ostfile );
        fatal( "could not write to owner state file" );
    }
    
    fclose( ostfile );

    info( "generated %s and %s", ostfilename, cstfilename );
    printout( "generated %s and %s\n", ostfilename, cstfilename );
        
	state_free( state );

    free( cstfilename );
    free( ostfilename );
}

int main( int argc, char** argv )
{
    int opt;
	char* pwdstr = NULL;
	char* logdir = NULL;
    char* resourcedir = NULL;
    char* statefilename = NULL;
    int encrypt_ostfile = TRUE; 

    while( ( opt = getopt( argc, argv, "p:l:r:s:po:h" ) ) != -1 ) {
        switch( opt ) {
            case 'p' : pwdstr = optarg; break;
            //case 'p' : encrypt_ostfile = TRUE; break;
	        case 'l' : logdir = optarg; break;
            case 'r' : resourcedir = optarg; break;
            case 's' : statefilename = optarg; break;
            case 'h' : /* fall through */
            default :
                usage( argv[0] );
                exit( EXIT_SUCCESS );
            break;
        }
    }
    
	setup_logdir( logdir, "adeona_init_log.txt" );
	setup_resources( resourcedir, RES_OPTIONS );
    opts.have_console = TRUE;

    int plen = 0;
	char pwdbuf[OST_PWDSTR_BUFLEN];
    if( encrypt_ostfile == TRUE ) {
        if( pwdstr == NULL ) {
			pwdstr = pwdbuf;
			plen = OST_PWDSTR_BUFLEN;
			pwdprompt_loop( pwdstr, &plen, 1 );
        
			if( plen == 0 ) {
				fatal( "No password given. Could not encrypt owner state file." );
			}
		}
		else {
			plen = strlen( pwdstr );
		}
    }
    
    do_init( statefilename, pwdstr, plen );

    exit( EXIT_SUCCESS );
}



