/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <getopt.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "common.h"
#include "options.h"
#include "log.h"
#include "util.h"
#include "core.h"
#include "remstore.h"

#define ARG_COUNT 2
#define TWO_DAYS  ( 60*60*24*2 )
#define FIVE_DAYS ( 60*60*24*5 )
#define OST_PWDSTR_BUFLEN 128

static inline void usage( char* prog )
{
    printf( "Usage %s [args] [options]\n"
            "Arguments:\n"
            "-s statefile.ost   \t" "Read owner state from statefile.ost\n"
            "Options:\n"
            "-p password        \t" "Specify password to decrypt owner statefile.\n"
            "-l logdir          \t"	"Directory where output logs and data are stored\n"
            "-r resourcedir     \t" "Directory containing resources (e.g. options.adeona)\n"
            "-t MM/DD/YYYY,HH:MM\t" "Lower bound on time for search, defaults to two days ago\n"
            "-e MM/DD/YYYY,HH:MM\t" "Upper bound on time for search, defaults to now\n"
            "-i key             \t" "Specify individual prgkey (in hex) to retrieve\n"
            "-n num             \t" "Retrieve last num locations, backwards from upper bound on time to no earlier than lower bound\n"
            "\n", prog );
}

static inline void print_disclaimer( void )
{
    printout( "*******************************************************************************\n"
              "* These results are for informational, research and evaluation purposes only. *\n"
              "* Do not attempt to recover your lost or stolen laptop yourself.              *\n"
              "* If you believe your laptop has been stolen, contact the appropriate         *\n"
              "* law enforcement agency.                                                     *\n"
              "*******************************************************************************\n\n" );
}

static inline void alloc_updatearray( update_t* updates, int size )
{
    int i;
    for( i = 0; i < size; i++ ) {
        updates[i].skipped = 0;
        location_alloc( &updates[i].loc );
    }
}

static inline void deepcopy_updatearray( update_t* dest, update_t* src, int size )
{
    int i;
    for( i = 0; i < size; i++ ) {
        dest[i].skipped = src[i].skipped;
        location_copy( &dest[i].loc, &src[i].loc );
    }
}

static inline void free_updatearray( update_t* updates, int size )
{
    int i;
    for( i = 0; i < size; i++ ) {
        updates[i].skipped = 0;
        location_free( &updates[i].loc );
    }
}

// Returns the number of distinct locations found
// Here we're trusting skipped to get us back to the correct state. 
static int decapsulate_all_and_return( state_t* state_init, state_t* state_temp, update_t updates[] )
{
    if( state_init != NULL && state_temp != NULL && updates != NULL ) {
        int numfound = 0;

        int j;
        for( j=0; j<opts.ctxt_queue_size; j++ ) {
            
            debug( "decapsulating cache element %d", j );
            log_msg_bin( ADEONA_CRYPTO, state_temp->currentseed, sizeof(prgkey_t), "decapsulating cache element %d using currentseed", j );
            log_msg_bin( ADEONA_CRYPTO, state_temp->cacheseed, sizeof(prgkey_t), "decapsulating cache element %d using cacheseed", j );

            if( update_decapsulate( state_temp ) == FAILURE ) {
                error( "could not decapsulate retrieved data" );
                return numfound; 
            }

            if( ctxt_cache_getelem( state_temp->cacheseed, state_temp->cache, &updates[j] ) == FAILURE ) {
                error( "could not decrypt cache element" );
                return numfound;
            }

            numfound++;

            info( "======== start location data =========" );
            log_msg_func( ADEONA_INFO, location_printdata_friendly, &updates[j].loc );
            info( "========= end location data ==========" );
            
            printout( "===============================\n" );
            printout( "Retrieved location information:\n" );
            location_printdata_friendly( stdout, &updates[j].loc );
            printout( "\n" );

            if( opts.use_webcamimg == TRUE ) {
                location_t* loc = &updates[j].loc;
                char filename[64];
                memset( filename, 0x00, sizeof( filename ) );
                snprintf( filename, sizeof( filename ), "adeona-snapshot-%s-%s.jpg", time2fstring( loc->timestamp, FALSE ), inet_ntoa( loc->extIP ) );
                
                char* savepath = NULL;
                if( STROPT_ISLOADED( opts.log_filenamebase ) ) {
                    savepath = concatstr( opts.log_filenamebase, filename ); 
                }
                else {
                    savepath = filename;
                }
                
                if( location_save_data( savepath, &updates[j].loc ) == SUCCESS ) {
                    info( "iSight snapshot saved in %s", savepath );
                    printout( "=> iSight snapshot saved in %s\n", savepath );
                }

                if( savepath != filename && savepath != NULL ) {
                    free( savepath );
                    savepath = NULL;
                }
            }
            
            printout( "\n" );
            printout( "===============================\n" );

            int backamount;
            if( ismsbset( updates[j].skipped ) == SUCCESS )
                backamount = 0;
            else 
                backamount = updates[j].skipped;

            if( backamount > 0 ) {
                if( sched_gobackn( state_init, state_temp, backamount ) == FAILURE ) {
                    error( "could not rewind state" );
                    return numfound;
                }
            }
        }

        return numfound;
    }
    else {
        error( "invalid input" );
        return 0;
    }
}

static int decapsulate_all( state_t* state_init, state_t* state_temp )
{
    update_t updates[opts.ctxt_queue_size];
    alloc_updatearray( updates, opts.ctxt_queue_size );
    int rv = decapsulate_all_and_return( state_init, state_temp, updates );
    free_updatearray( updates, opts.ctxt_queue_size );

    return rv; 
}

/*
 * Realized this chunk of code is all over the place so wrapped it
 * here.
 * 
 * <i> is the ith DHT key from replication.
 *
 */ 
static int do_data_retrieve( state_t* state, int i )
{
    assert( state != NULL && i >= 0 );
    
    if( remstore_isconnected() == FALSE ) {
        debug( "connection was lost, reconnecting..." );
        printout( "Connection was lost, reconnecting\n" );
        if( remstore_connect() == FAILURE ) {
            error( "could not reconnect" );
            printout( "Could not reconnect to remote storage server\n" );
            return FAILURE;
        }
    }

    log_msg_bin( ADEONA_DEBUG, state->sd->dhtkeys[i], sizeof( dhtkey_t ), "GET from dht key %d", i );
    
    getiter_t iter;

    if( remstore_getiter_init( state->sd->dhtkeys[i], &iter ) == FAILURE ) {
        info( "update replica %d not available", i );
        printout( "Update replica %d not available\n", i );
        return FAILURE;
    }
    
    if( remstore_getiter_next( &iter, (uchar*)state->cache, CTXT_QUEUE_LEN ) == FAILURE ) {
        info( "update replica %d not available", i );
        printout( "Update replica %d not available\n", i );
        remstore_getiter_done( &iter );
        return FAILURE;
    }
    
    remstore_getiter_done( &iter );
    
    log_msg_bin( ADEONA_DEBUG, (uchar*)state->cache, CTXT_QUEUE_LEN, "retrieved ctxtcache:" );
    
    info( "succesfully retrieved update replica %d", i );
    printout( "Succesfully retrieved update replica %d\n", i );
    
    return SUCCESS;
}

static int retrieve_locupdate( state_t* state_init, state_t* state, update_t updates[] )
{
    int i;
    int maxRecovered = 0;
    state_t* state_temp = state_alloc();

    update_t temp_updates[opts.ctxt_queue_size];
    alloc_updatearray( temp_updates, opts.ctxt_queue_size );
    

    for( i=0; i<opts.num_dhtkeys_totry; i++ ) {
        
        if( do_data_retrieve( state, i ) == FAILURE ) {
            continue;
        }
 
        state_copy( state_temp, state, SC_ALL );
        
        int numRecovered = decapsulate_all_and_return( state_init, state_temp, temp_updates );

        if( numRecovered > opts.ctxt_queue_size ) {
            warn( "found %d locations when max is %d", numRecovered, opts.ctxt_queue_size );
            continue;
        }
        if( numRecovered > maxRecovered ) {
            deepcopy_updatearray( updates, temp_updates, opts.ctxt_queue_size );
            maxRecovered = numRecovered; 
        }
        if( numRecovered == opts.ctxt_queue_size )
            break; // found all we will find
    }
    
    free_updatearray( temp_updates, opts.ctxt_queue_size );
    state_free( state_temp );

    debug( "found %d", maxRecovered );
    return maxRecovered;
}


static void retrieve_nonpanic_single( prgkey_t targetseed, state_t* state_init )
{
    if( state_init == NULL ) {
        fatal( "invalid input" );
    }

    state_t* state_curr = state_alloc();
    state_t* state_temp = state_alloc();
       
    state_copy( state_curr, state_init, SC_NOCACHE );
    
    int i = 0;
    debug( "looking for target seed..." );
    while( memcmp( state_curr->currentseed, targetseed, sizeof(prgkey_t) ) != 0  ) {
        i++;
        sched_nextstate( state_curr );
        if( i % 50 == 0 ) { 
            debug( "%d states explored so far...", i );
        }
    }
    debug( "target seed found after %d iterations", i );

    if( remstore_connect() == FAILURE ) {
        fatal( "could not connect to remote storage server, try again" );
    }
    
    debug( "========== START STATE RETRIEVE ==========" );
    log_msg_bin( ADEONA_CRYPTO, state_curr->currentseed, sizeof( prgkey_t ), "state currentseed:" );
    debug( "state updatetime: %s", time2string( state_curr->updatetime, FALSE ) );
        
    for( i=0; i<opts.num_dhtkeys_totry; i++ ) {
        
        if( do_data_retrieve( state_curr, i ) == FAILURE ) {
            continue;
        }
            
        state_copy( state_temp, state_curr, SC_ALL );

        decapsulate_all( state_init, state_temp );
    }
    
    debug( "==========  END STATE RETRIEVE  ==========" );

    remstore_disconnect();
    
    state_free( state_curr );
    state_free( state_temp );

}

// Attempts to retrieve the last n distinct location updates, starting
// at endtime and going backwards to no earlier than starttime
static void retrieve_nonpanic_lastn( int n, time_t starttime, time_t endtime, state_t* state_init )
{
    if( state_init == NULL ) {
        fatal( "invalid input" );
    }

    state_t* state_curr = state_alloc();
    state_t* state_temp = state_alloc();
       
    state_copy( state_curr, state_init, SC_NOCACHE );
    
    // Find first state to retrieve
    if( sched_seektotime( state_curr, endtime ) == FAILURE ) {
        fatal( "could not update state" );
    }
    
	if( memcmp( state_init->currentseed, state_curr->currentseed, sizeof( prgkey_t ) ) == 0 ) {
		info( "no updates to retrieve from specified time period" );
		printout( "\nNo updates to retrieve from specified time period.\n" );
		return;
    }

    if( remstore_connect() == FAILURE ) {
        fatal( "could not connect to remote storage server, try again" );
    }
    
    int numfound = 0; 
    location_t lastloc_seen;
    memset( &lastloc_seen, 0, sizeof(location_t) ); // Not using trace route data here

    info( "========== START STATE RETRIEVE ==========" );
    while( numfound < n && state_curr->updatetime >= starttime ) { 
        update_t updates[opts.ctxt_queue_size];
        memset( updates, 0, sizeof(update_t) * opts.ctxt_queue_size );
        alloc_updatearray( updates, opts.ctxt_queue_size );  

        log_msg_bin( ADEONA_CRYPTO, state_curr->currentseed, sizeof( prgkey_t ), "state currentseed:" );
        info( "state updatetime: %s", time2string( state_curr->updatetime, FALSE ) );
        printout( "\nChecking update scheduled on %s\n", time2string( state_curr->updatetime, FALSE ) ); 

        int numRecovered = retrieve_locupdate( state_init, state_curr, updates );
                
        int j;
        int alreadySeen = 0;
        for( j=0; j < numRecovered; j++ ) {
            
            // Report these locations 
            if( location_compare( &updates[j].loc, &lastloc_seen ) == SUCCESS ) {
                alreadySeen++;
                continue;
            }
            
            debug( "*** %d %d", numfound, updates[j].loc.timestamp );
            debug( "*** %d %s", numfound, inet_ntoa( updates[j].loc.intIP ) );
            debug( "*** %d %s", numfound, inet_ntoa( updates[j].loc.extIP ) );

            lastloc_seen = updates[j].loc;
        }
        
        numfound += numRecovered - alreadySeen;
        
        uint skip_amount = 1;
        if( numRecovered > 0 ) {
            skip_amount = updates[0].skipped;
            if( ismsbset( skip_amount ) ) 
                skip_amount = flipmsb( updates[0].skipped );

            if( skip_amount <= 0 ) {
                warn( "skip amount too small" );
                skip_amount = 1;
            }
        }

        if( sched_gobackn( state_init, state_curr, skip_amount ) == FAILURE ) {
            error( "could not rewind state further" );
            break;
        }
		
		if( memcmp( state_init->currentseed, state_curr->currentseed, sizeof( prgkey_t ) ) == 0 ) {
			break;
		}
    
        free_updatearray( updates, opts.ctxt_queue_size );

    }//end while
    info( "==========  END STATE RETRIEVE  ==========" );
    
    remstore_disconnect();
    
    state_free( state_curr );
    state_free( state_temp );
}


static void retrieve_nonpanic( time_t starttime, time_t endtime, state_t* state_init )
{
    if( state_init == NULL ) {
        fatal( "invalid input" );
    }
    state_t* state_curr = state_alloc();
    state_t* state_temp = state_alloc();
       
    state_copy( state_curr, state_init, SC_NOCACHE );
    
    // Find first state to retrieve
    if( sched_seektotime( state_curr, endtime ) == FAILURE ) {
        fatal( "could not update state" );
    }
    
	if( memcmp( state_init->currentseed, state_curr->currentseed, sizeof( prgkey_t ) ) == 0 ) {
		info( "no updates to retrieve from specified time period." );
		printout( "\nNo updates to retrieve from specified time period.\n" );
		return;
    }

    if( remstore_connect() == FAILURE ) {
        fatal( "could not connect to remote storage server, try again" );
    }
    
    while( state_curr->updatetime >= starttime ) { 

        info( "========== START STATE RETRIEVE ==========" );
        log_msg_bin( ADEONA_CRYPTO, state_curr->currentseed, sizeof( prgkey_t ), "state currentseed:" );
        info( "state updatetime: %s", time2string( state_curr->updatetime, FALSE ) );
        printout( "\nChecking update scheduled on %s\n", time2string( state_curr->updatetime, FALSE ) ); 
        
        int i;
        for( i=0; i<opts.num_dhtkeys_totry; i++ ) {
            
            if( do_data_retrieve( state_curr, i ) == FAILURE ) {
                continue;
            }
                
            state_copy( state_temp, state_curr, SC_ALL );

            decapsulate_all( state_init, state_temp );
        }
        
        info( "==========  END STATE RETRIEVE  ==========" );
        
        if( sched_prevstate( state_init, state_curr ) == FAILURE ) {
            error( "could not get previous state" );
        }

        if( memcmp( state_init->currentseed, state_curr->currentseed, sizeof( prgkey_t ) ) == 0 ) {
            break;
        }

    }//end while
    
    remstore_disconnect();
    
    state_free( state_curr );
    state_free( state_temp );
}

int main( int argc, char** argv )
{
    if( argc < ARG_COUNT ) {
        usage( argv[0] );
        exit( EXIT_FAILURE );
    }
    
    int opt;
    char* pwdstr = NULL;
    char* logdir = NULL;
    char* resourcedir = NULL;
    char* ostatefilename = NULL;
    FILE* ostatefile = NULL;
    char* starttimestr = NULL;
    char* endtimestr = NULL;    
    char* currentseedstr  = NULL;
    int num = -1;

    prgkey_t targetseed;
    
    time_t starttime = (time_t) 0;
    time_t endtime = (time_t) 0;

	// NOTE: defaulting to always using encryption.
    int encrypted_ostfile = TRUE;

    while( ( opt = getopt( argc, argv, "s:p:l:r:t:e:i:n:h" ) ) != -1 ) {
        switch( opt ) {
            case 's' : ostatefilename = optarg; break;
            //case 'p' : encrypted_ostfile = TRUE; break;
            case 'p' : pwdstr = optarg; break; //printf( "pwdstr is %s\n", pwdstr ); break;
            case 'l' : logdir = optarg; break; //printf( "logdir is %s\n", logdir );break;
            case 'r' : resourcedir = optarg; break; //printf( "resourcedir is %s\n", resourcedir );break;
            case 't' : starttimestr = optarg; break; //printf( "starttimestrstr is %s\n", starttimestr );break;
            case 'e' : endtimestr = optarg; break; //printf( "endtimestr is %s\n", endtimestr );break;
            case 'i' : currentseedstr = optarg; break;
            case 'n' : num = atoi( optarg ); break; //printf( "n is %d\n", num );break;
            case 'h' : /* fall through */
            default :
                usage( argv[0] );
                exit( EXIT_SUCCESS );
            break;
        }
    }

    if( ostatefilename == NULL ) {  
        usage( argv[0] );
        exit( EXIT_FAILURE );
    }

	setup_logdir( logdir, "adeona_retrieve_results.txt" );
	setup_resources( resourcedir, RES_OPTIONS | RES_GATEWAYS );
    opts.have_console = TRUE;

    if( ( ostatefile = fopen( ostatefilename, "r" ) ) == NULL ) {
		error( "could not open statefile %s", ostatefilename );
		return FAILURE;
	}

    // NOTE: If options in ostatefile state new log file or gateway we'll be reporting
    // errors. Right now we're not specifying
    // these options during initialize in ost files, so it's a mute point unless a user
    // manually edits options in the ost file.
    parse_options_file( ostatefile );

   	state_t* state_init = state_alloc();
    
    if( encrypted_ostfile == TRUE ) {
        int plen = 0;
        char pwdbuf[OST_PWDSTR_BUFLEN];
        
        if( pwdstr == NULL ) {
			pwdstr = pwdbuf;
			plen = OST_PWDSTR_BUFLEN;
			pwdprompt_loop( pwdstr, &plen, 0 );
        
			if( plen == 0 ) {
				fatal( "No password given" );
			}
		}
	    else {
            plen = strlen( pwdstr );
        }
		
        int loadstatus = FAILURE;
		int attempt = 0;
		int statefilepos = ftell( ostatefile ); 
		while( ( loadstatus = state_load( ostatefile, state_init, pwdstr, plen, SC_NOCACHE ) ) == FAILURE 
					&& attempt++ < 3 ) 
        {
			pwdstr = pwdbuf;
			plen = OST_PWDSTR_BUFLEN;
			pwdprompt_loop( pwdstr, &plen, 0 );
        
			if( plen == 0 ) {
				printout( "No password given" );
			}

			fseek( ostatefile, statefilepos, SEEK_SET );
		}

		if( loadstatus == FAILURE ) {
			fatal( "could not load state from %s", ostatefilename ); 
		}
    }
    
    fclose( ostatefile );

    // Do a retrieve for a particular state, identified by currentseed
    if( currentseedstr != NULL ) {
        if( strlen(currentseedstr) != PRG_HEX_KEYSIZE ) {
            fatal( "current seed specified on command line not correct length, should be %d", PRG_HEX_KEYSIZE );
        }
        hex2bin( (uchar*)currentseedstr, PRG_HEX_KEYSIZE, (uchar*)targetseed, sizeof(prgkey_t) );

        retrieve_nonpanic_single( targetseed, state_init );
    }
    else {
        int isdefault_endtime = FALSE;
        int isdefault_starttime = FALSE;

        // Determine end time for search
        if( endtimestr != NULL  ) {
            if( ( endtime = string2time( endtimestr ) ) == (time_t)-1 ) {
                warn( "end time given but could not be parsed, using default" );
                endtime = time( NULL );
                isdefault_endtime = TRUE;
            }
        }
        else {
            endtime = time( NULL ); // if none just use current time
            isdefault_endtime = TRUE;
        }

        // Determine start time for search
        if( starttimestr != NULL ) {
            if( ( starttime = string2time( starttimestr ) ) == (time_t)-1 ) {
                warn( "start time given but could not be parsed, using default" );
                starttime = endtime - TWO_DAYS;
                isdefault_starttime = TRUE;
            }
        }
        else {
            starttime = endtime - TWO_DAYS;  // if none just use 2 days before endtime
            isdefault_starttime = TRUE;
        }
    
        // Time sanity check
        if( starttime > endtime ) { 
            fatal( "start time is greater than end time" );
        }
        
        print_disclaimer();

        info( "start time %s", time2string( starttime, FALSE ) );
        info( "end time   %s", time2string( endtime, FALSE ) );

        if( num > -1 ) {
            // Do a search starting at end time and finding first num locations
            // or stopping at starttime (whichever happens first).
            printout( "\nSearching for most recent %d update(s) in time period ", num );
            printout( "[ %s - ", time2string( starttime, FALSE ) );
            if( isdefault_endtime == TRUE ) {
                printout( "NOW ]\n\n" );
            }
            else {
                printout( "%s ]\n\n", time2string( endtime, FALSE ) );
            }

            retrieve_nonpanic_lastn( num, starttime, endtime, state_init );
        }
        else {
            // Do a search starting from starttime and ending at endtime
            printout( "\nRetrieving *all* updates for time period " );
            printout( "[ %s - ", time2string( starttime, FALSE ) );
            if( isdefault_endtime == TRUE ) {
                printout( "NOW ]\n\n" );
            }
            else {
                printout( "%s ]\n\n", time2string( endtime, FALSE ) );
            }

            retrieve_nonpanic( starttime, endtime, state_init );
        }
    }
    
    //interesting...
    //printf( "start: %s, end: %s\n", time2string( starttime ), time2string( endtime ) );
    
    
    exit( EXIT_SUCCESS );
}

