/*
 * Copyright (c) 2007-2008 University of Washington
* 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <unistd.h>

#include "state.h"
#include "log.h"
#include "util.h"

typedef struct _encrypted_state_keys
{
    enckey_t enckey;
    hmackey_t hmackey;
} encstatekey_t;

state_t* state_alloc( void )
{
    state_t* state = (state_t*)malloc( sizeof( state_t ) );
    assert( state != NULL );
    memset( state, 0x00, sizeof( state_t ) );

    state->cache = ctxt_cache_alloc();
    assert( state->cache != NULL );

    state->sd = (struct state_data*)malloc( sizeof( struct state_data ) );
    assert( state->sd != NULL );
    memset( state->sd, 0x00, sizeof( struct state_data ) );

    state->sd->dhtkeys = (dhtkey_t*)malloc( sizeof( dhtkey_t ) * opts.num_dhtkeys_totry );
    assert( state->sd->dhtkeys  != NULL );
    memset( state->sd->dhtkeys, 0x00, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry );
    
    return state;
}

/*
 * File opening wrapper for state_save
 * 
 */ 
int state_save_filename( char* filename IN, state_t* state IN, char* pwdstr IN, int plen IN, uint flags IN )
{
    FILE* file;

    if( (file = fopen( filename, "w" ) ) == NULL ) {
        error( "could not open statefile for write" );
        return FAILURE;
    }

    int rv = state_save( file, state, pwdstr, plen, flags );

    fclose( file );
    return rv;
}

/*
 * Save state_t to file.
 * 
 * If <pwdstr> is provided, SC_SAVED is encrypted under it.
 * SC_CACHE is ciphertext itself so there is no need to encrypt it.
 *
 */ 
int state_save( FILE* file IN, state_t* state IN, char* pwdstr IN, int plen IN, uint flags IN ) 
{
    if( file != NULL && state != NULL ) {   
        if( flags & SC_SAVED ) {
            //TODO: pretty code needed below
            if( pwdstr != NULL && plen != 0 ) {
                pbkdfsalt_t salt;
                encstatekey_t pbkeys;

                if( get_random_bytes( salt, sizeof( pbkdfsalt_t ) ) == FAILURE ) {
                    error( "could not generate salt" );
                    fclose( file );
                    return FAILURE;
                }
                
                //generate a key using pbkdf then that key is split in
                //two via the authenckey_t type i.e. generate a 256
                //bit key and use the first 128 bits for encryption
                //and the last half for hmac
                if( pbkdf_generate( pwdstr, plen, salt, (uchar*)&pbkeys, sizeof( encstatekey_t ) ) == FAILURE ) {
                    error( "could not generate key from password" );
                    fclose( file );
                    return FAILURE;
                }
                
                int slen = SC_SAVED_SIZE + HMAC_LEN;
                uchar state_ctxt[slen];
                uchar* sptr = state_ctxt;
                
                //serialize
                datacpy( sptr, &state->updatetime, sizeof( time_t ) );
                sptr += sizeof( time_t );
                memcpy( sptr, state->currentseed, sizeof( prgkey_t ) );
                sptr += sizeof( prgkey_t );
                memcpy( sptr, state->cacheseed, sizeof( prgkey_t ) );

                //encrypt the buffer
                if( encrypt_then_mac( pbkeys.enckey, pbkeys.hmackey, state_ctxt, SC_SAVED_SIZE,  state_ctxt, slen ) == FAILURE ) {
                    memset( state_ctxt, 0x00, slen );
                    memset( &pbkeys, 0x00, sizeof( encstatekey_t ) );
                    fclose( file );
                    return FAILURE;
                }
                
                //save the salt and the ciphertext
                if( fprintln_bin2hex( file, salt, sizeof( pbkdfsalt_t ) ) == FAILURE
                    || fprintln_bin2hex( file, state_ctxt, slen ) == FAILURE )
                {
                    error( "could not write salt and ciphertext to file" );
                    memset( state_ctxt, 0x00, slen );
                    memset( &pbkeys, 0x00, sizeof( encstatekey_t ) );
                    return FAILURE;
                }
                
                memset( state_ctxt, 0x00, slen );
                memset( &pbkeys, 0x00, sizeof( encstatekey_t ) );
            }
            else {
                if( fprintln_data2hex( file, &state->updatetime, sizeof( time_t ) ) == FAILURE 
                    || fprintln_bin2hex( file, state->currentseed, sizeof( prgkey_t ) ) == FAILURE
                    || fprintln_bin2hex( file, state->cacheseed, sizeof( prgkey_t ) ) == FAILURE )
                {
                    error( "could not write state to file" );
                    return FAILURE;
                }
            }
        }
        
        if( flags & SC_CACHE ) {
            if( ctxt_cache_save( file, state->cache ) == FAILURE ) {
                error( "could not save the cache" );
                fflush( file );
                fclose( file );
                return FAILURE;
            }
        }
        
        fflush( file );
        fsync( fileno( file ) );

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }

}


/*
 * File-loading wrapper for state_load
 *
 */
int state_load_filename( char* filename IN, state_t* state INOUT, char* pwdstr IN, int plen IN, uint flags IN )
{
    FILE* file = NULL;

    if( ( file = fopen( filename, "r" ) ) == NULL ) {
		error( "could not open statefile %s", filename );
		return FAILURE;
	}	


    int rv = state_load( file, state, pwdstr, plen, flags );
    fclose( file );
    return rv;
}

/*
 * Read in from a state_t file as created by state_save.
 *
 */
int state_load( FILE* statefile IN, state_t* state INOUT, char* pwdstr IN, int plen IN, uint flags IN )
{
    if( statefile != NULL && state != NULL ) {

		//if( ( file = fopen( filename, "r" ) ) == NULL ) {
		//	error( "could not open statefile %s", filename );
		//	return FAILURE;
		//}

        if( flags & SC_SAVED ) {
            if( pwdstr != NULL && plen > 0 ) {
                pbkdfsalt_t salt;
                encstatekey_t pbkeys;

                int slen = SC_SAVED_SIZE + HMAC_LEN;
                uchar state_ctxt[slen];
                uchar* sptr = state_ctxt;
                
                if( freadln_hex2bin( statefile, salt, sizeof( pbkdfsalt_t ) ) == FAILURE 
                    || freadln_hex2bin( statefile, state_ctxt, slen ) == FAILURE )
                {
                    error( "could not read from state file" );
                    return FAILURE;
                }
                
                if( pbkdf_generate( pwdstr, plen, salt, (uchar*)&pbkeys, sizeof( encstatekey_t ) ) == FAILURE ) {
                    error( "could not generate key from password" );
                    //fclose( statefile );
                    return FAILURE;
                }
                
                if( verify_then_decrypt( pbkeys.enckey, pbkeys.hmackey, state_ctxt, slen, state_ctxt, SC_SAVED_SIZE ) == FAILURE ) {
                    error( "could not decrypt state ciphertext" );
                    memset( state_ctxt, 0x00, slen );
                    memset( &pbkeys, 0x00, sizeof( encstatekey_t ) );
                    //fclose( file );

                    // NOTE: Tom added this as feedback to user, since we were failing silently
                    printout( "Password failed.\n\n" );
                    return FAILURE;
                }
                
                //deserialize
                datacpy( &state->updatetime, sptr, sizeof( time_t ) );
                sptr += sizeof( time_t );
                memcpy( state->currentseed, sptr, sizeof( prgkey_t ) );
                sptr += sizeof( prgkey_t );
                memcpy( state->cacheseed, sptr, sizeof( prgkey_t ) );
                
                memset( state_ctxt, 0x00, slen );
                memset( &pbkeys, 0x00, sizeof( encstatekey_t ) );
            }
            else {
                if( freadln_hex2data( statefile, &state->updatetime, sizeof( time_t ) ) == FAILURE 
                    || freadln_hex2bin( statefile, state->currentseed, sizeof( prgkey_t ) ) == FAILURE 
                    || freadln_hex2bin( statefile, state->cacheseed, sizeof( prgkey_t ) ) == FAILURE )
                {
                    error( "could not read statefile" );
                    return FAILURE;
                }
            }
        }

        if( flags & SC_CACHE ) {
            if( ctxt_cache_load( statefile, state->cache ) == FAILURE ) {
                //fclose( statefile );
                return FAILURE;
            }
        }
        
        // we need SC_SAVED to load SC_DATA
        if( ( flags & SC_DATA ) && ( flags & SC_SAVED ) ) {
            if( prg_generate( state->currentseed, STATE_NEXTSEED_LOW, STATE_RENC_HIGH, (uchar*)state->sd ) == FAILURE 
                || prg_generate( state->currentseed, STATE_DHT_LOW, STATE_DHT_HIGH, (uchar*)state->sd->dhtkeys ) == FAILURE ) 
            {
                //fclose( statefile );
                error( "could not generate SC_DATA" );
                return FAILURE;
            }
        }

		//fclose( statefile );
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Only the init program uses this.
 *
 */ 
int state_makenew( state_t* state OUT )
{
    if( state != NULL ) {
        time( &state->updatetime );
        if( state->updatetime == (time_t)-1 ) {
            error( "could not get time" );
            return FAILURE;
        }
        
        if( get_random_bytes( state->currentseed, sizeof( prgkey_t ) ) == FAILURE ) {
            error( "could not generate new main prg key" );
            return FAILURE;
        }

		if( get_random_bytes( state->cacheseed, sizeof( prgkey_t ) ) == FAILURE ) {
            error( "could not generate new cache prg key" );
            return FAILURE;
        }

		if( ctxt_cache_fillrand( state->cache ) == FAILURE ) {
			error( "could not initialize cache" );
			return FAILURE;
		}
        
        //leave state->sd alone, init program does not to generate it

        return SUCCESS;   
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }

}

/*
 * Copy parts of state if the flag for it is set.
 * 
 */ 
int state_copy( state_t* dst INOUT, state_t* src IN, uint flags IN )
{
    if( dst != NULL && src != NULL ) { 
        
        if( flags & SC_SAVED ) {
            memcpy( &dst->updatetime, &src->updatetime, sizeof( time_t ) );
            memcpy( dst->currentseed, src->currentseed, sizeof( prgkey_t ) );
            memcpy( dst->cacheseed, src->cacheseed, sizeof( prgkey_t ) );
        }

        if( flags & SC_CACHE ) {
            if( dst->cache != NULL && src->cache != NULL ) {
                memcpy( dst->cache, src->cache, CTXT_QUEUE_LEN );
            }
            else {
                error( "no state cache" );
                return FAILURE;
            }
        }
        
        if( flags & SC_DATA ) {
            if( dst->sd != NULL && src->sd != NULL ) {
                memcpy( dst->sd->nextseed, src->sd->nextseed, sizeof( prgkey_t ) );
                memcpy( dst->sd->nextcacheseed, src->sd->nextcacheseed, sizeof( prgkey_t ) );
                memcpy( dst->sd->schedseed, src->sd->schedseed, sizeof( prgblock_t ) );
                memcpy( dst->sd->renckey, src->sd->renckey, sizeof( renckey_t ) );  
                memcpy( dst->sd->dhtkeys, src->sd->dhtkeys, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry );  
            }
            else {
                error( "no state data" );
                return FAILURE;
            }
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}   

/*
 * Deep free of state_t
 *
 */
void state_free( state_t* state IN )
{
    if( state != NULL ) {
        
        ctxt_cache_del( state->cache );

        if( state->sd != NULL ) {

            if( state->sd->dhtkeys != NULL ) {
                memset( state->sd->dhtkeys, 0x00, sizeof( dhtkey_t ) * opts.num_dhtkeys_totry ); 
                free( state->sd->dhtkeys );
                state->sd->dhtkeys = NULL;
            }
            
            memset( state->sd, 0x00, sizeof( struct state_data ) );
            free( state->sd );
            state->sd = NULL;
        }
        
        memset( state, 0x00, sizeof( state_t ) );
        free( state );
    }
    else {
        error( "invalid input" );
    }
}


