/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */

#include "update.h"
#include "sched.h"
#include "crypto.h"
#include "log.h"

/*
 * Prepares cache for being sent to remote storage, i.e. by (re)encrypting it. 
 *
 * This reencryption ensures that, even if cache elements have not changed locally,
 * the remote storage will always just see random bits. 
 *
 */ 
int update_encapsulate( state_t* state INOUT )
{
	if( state != NULL && state->sd != NULL ) {

        if( encryption( state->sd->renckey,
                        (uchar*)state->cache,
                        (uchar*)state->cache, 
                        CTXT_QUEUE_LEN ) == FAILURE ) 
        {
            error( "could not encrypt cache" );
            return FAILURE;
        }
         
		return SUCCESS;
	}
	else {
		error( "invalid input" );
		return FAILURE;
	}
}

int update_decapsulate( state_t* state INOUT )
{
	if( state != NULL && state->sd != NULL ) {
 
        if( decryption( state->sd->renckey,
                        (uchar*)state->cache,
                        (uchar*)state->cache,
                        CTXT_QUEUE_LEN ) == FAILURE ) 
        {
            error( "could not decrypt cache" );
            return FAILURE;
        }
            
        return SUCCESS;       
        
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}
 
/*
 * IMPORTANT: Saves next state back out to disk in case update process
 * gets interrupted. Should be called before remote storage attempts.
 * 
 */
int update_interrupt_protection( char* statefilename, state_t* state IN )
{
    if( statefilename != NULL && state != NULL ) {
        state_t* nextstate = state_alloc(); //TODO: keep this around so we don't call allocator everytime
        assert( nextstate != NULL );

        if( state_copy( nextstate, state, SC_ALL ) == FAILURE ) {
            error( "could not copy current state" );
            return FAILURE;
        }
        
        if( sched_nextstate( nextstate ) == FAILURE ) {
            error( "could not move state forward" );
            return FAILURE;
        }
        
		if( state_save_filename( statefilename, nextstate, NULL, 0, SC_ALL ) == FAILURE ) {
			error( "could not save updated state" );
			return FAILURE;
		}
        
        state_free( nextstate );

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

