/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <arpa/inet.h>

#include "pbkdf.h"
#include "hmac.h"
#include "util.h"
#include "log.h"

#define PBKDF_HMACSHA1_ITERS 1000 //recommended in PKCS#5 v2.0

/*
 * The xor sum of the first <c> iterations of HMAC-SHA1
 *
 */ 
static void xorsum_iterations( char* P, int plen, uchar* S, int slen, int c, int i, uchar* T, int tlen )
{
    unsigned int index = htonl( i ); // most significant octet first
    uchar scati[slen + sizeof( unsigned int )]; // ( S||INT(i) )
    memcpy( scati, S, slen );
    memcpy( &scati[slen], &index, sizeof( unsigned int ) );

    uchar Ucur[HMACSHA1_LEN];
    uchar Uprev[HMACSHA1_LEN];
    memset( Ucur, 0x00, HMACSHA1_LEN );
    memset( Uprev, 0x00, HMACSHA1_LEN );
    
    hmacsha1_compute( (uchar*) P, plen, scati, slen+sizeof( unsigned int ), Uprev );
    assert( tlen <= HMACSHA1_LEN );
    memcpy( T, Uprev, tlen );

    while( c-- > 1 ) {
        hmacsha1_compute( (uchar*)P, plen, Uprev, HMACSHA1_LEN, Ucur );
        
        int j;
        for( j=0; j<tlen; j++ ) {
            T[j] ^= Ucur[j];
        }
        memcpy( Uprev, Ucur, HMACSHA1_LEN );
    }

    memset( Ucur, 0x00, HMACSHA1_LEN );
    memset( Uprev, 0x00, HMACSHA1_LEN );
}

/*
 * Get an encryption key <dkey> from a string <password>. Must provide
 * a <salt> value.
 *
 */ 
int pbkdf_generate( char* password IN, int plen IN, pbkdfsalt_t salt IN, uchar* dkey OUT, int dklen IN )
{
    return pbkdf2_hmacsha1( password, plen, salt, sizeof( pbkdfsalt_t ), PBKDF_HMACSHA1_ITERS, dkey, dklen );
}

/*
 * Implement PBKDF2 of PKCS#5 v2.0 with HMAC-SHA1
 * 
 */ 
int pbkdf2_hmacsha1( char* password IN, int plen IN, uchar* salt IN, int slen IN, int count IN, uchar* dkey OUT, int dklen IN )
{
    if( password != NULL && salt != NULL && dkey != NULL && plen > 0 && slen > 0 && dklen > 0 ) {
        
        if( dklen > INT_MAX ) {
            error( "derived key too long" );
            return FAILURE;
        }
        
        int nblocks   = dklen / HMACSHA1_LEN;
        int remainder = dklen % HMACSHA1_LEN;

        if( remainder > 0 ) {
            nblocks += 1;
        }
        
        assert( nblocks > 0 );

        int i, offset;
        for( i=1; i<nblocks; i++ ) {
            offset = ( i - 1 ) * HMACSHA1_LEN;
            xorsum_iterations( password, plen, salt, slen, count, i, &dkey[offset], HMACSHA1_LEN );
        }

        //last block
        i      = nblocks;
        offset = dklen - remainder;
        xorsum_iterations( password, plen, salt, slen, count, nblocks, &dkey[offset], remainder );
        
        return SUCCESS;
      
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}


