/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <openssl/crypto.h>
#include <openssl/evp.h>

#include "prg.h"
#include "log.h"
#include "util.h"

/* 
 * This function takes as input <prgkey>, already initialized, and a
 * range of bytes to generate. It then generates these byte offsets.
 * Note that if the byte offsets might fall over a variety of blocks
 * of prg output, in which case we need call AES several times. But if
 * they're just in one block, we need only call AES once. In general
 * we'll be calling AES minimal number of times.
 *
 * The caller MUST already have allocated a <outbuf> of the
 * appropriate size.
 *
 */
int prg_generate( prgkey_t prgkey IN, int rangeLow IN, int rangeHigh IN, uchar* outbuf OUT )
{
    if( prgkey != NULL && outbuf != NULL && ( rangeLow <= rangeHigh ) ) {
        EVP_CIPHER_CTX ctx;
        int firstBlock, lastBlock, lowMod, highMod;
        int i, ctlen, numbytes, bufctr, buflen;
        prgblock_t counter;
        prgblock_t stream;
        
        EVP_CIPHER_CTX_init( &ctx );
        EVP_CIPHER_CTX_set_padding( &ctx, 0 ); //disable padding
        
        //note to self: EVP routines with no "ex" when available are deprecated
        //NULL IV because ECB mode
        if( !EVP_EncryptInit_ex( &ctx, PRG_BLOCKCIPHER, NULL, prgkey, NULL ) ) {
            error( "cannot initialize" );
            return FAILURE;
        }

        // determine which blocks of output to generate
        firstBlock = rangeLow / PRG_BLOCKSIZE;
        lastBlock = rangeHigh / PRG_BLOCKSIZE;
        lowMod = rangeLow % PRG_BLOCKSIZE;
        highMod = rangeHigh % PRG_BLOCKSIZE;
        buflen = rangeHigh - rangeLow + 1;
        
        if( firstBlock >= PRG_MAX_COUNTER || lastBlock > PRG_MAX_COUNTER ) {
            error( "invalid byte range" );
            return FAILURE;
        }
        
        //some asserts for EVP_EncryptUpdate, input size is always the same as block size
        assert( sizeof( prgblock_t ) == PRG_BLOCKSIZE );
        assert( EVP_CIPHER_block_size( PRG_BLOCKCIPHER ) == PRG_BLOCKSIZE );
        
        // generate keystream
        bufctr = 0;
        memset( counter, 0x00, PRG_BLOCKSIZE ); //clear counter
        for( i=firstBlock; i<=lastBlock; i++ ) {
            datacpy( counter, (uchar*)&i, sizeof( int ) );
            
            ctlen = 0;
            if( !EVP_EncryptUpdate( &ctx, stream, &ctlen, counter, PRG_BLOCKSIZE ) ) {
                error( "could not encrypt counter" );
                return FAILURE;
            }
            
            if( ctlen != PRG_BLOCKSIZE ) {
                error( "could not encrypt all of counter" );
                memset( outbuf, 0x00, buflen );
                return FAILURE;
            }
            
            //now throw the appropriate bytes into the output buffer
            if( i == firstBlock ) {
                if( i == lastBlock ) {
                    numbytes = rangeHigh - rangeLow + 1;
                }
                else {
                    numbytes = PRG_BLOCKSIZE - lowMod;
                }
                
                assert( bufctr + numbytes <= buflen );
                memcpy( &outbuf[bufctr], &stream[lowMod], numbytes );
                bufctr += numbytes;
            }
            else if( i == lastBlock ) {
                // last block, but not first block. Just copy the
                // remaining number of bytes, starting at beginning of
                // streamblock. Don't need to increment bufptr
                numbytes = rangeHigh - lastBlock*PRG_BLOCKSIZE + 1;
                assert( bufctr + numbytes <= buflen );
                memcpy( &outbuf[bufctr], stream, numbytes );
                bufctr += numbytes;
            }
            else {
                // somewhere in middle, copy whole block over
                numbytes = PRG_BLOCKSIZE;
                assert( bufctr + numbytes <= buflen );
                memcpy( &outbuf[bufctr], stream, numbytes );
                bufctr += PRG_BLOCKSIZE;
            }
        }   
        
        EVP_CIPHER_CTX_cleanup( &ctx );
        assert( bufctr == buflen );
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

