/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#if defined( __CYGWIN__ )
#include <windows.h>
#endif

#include "log.h"
#include "options.h"
#include "util.h"
#include "accesspoint.h"
#include "config.h"

#if defined( __APPLE__ )
#define AP_WIPROG_CMDSTR "/System/Library/PrivateFrameworks/Apple80211.framework/Versions/Current/Resources/airport -I"
#endif

#if defined( __CYGWIN__ )
#define AP_WIPROG_CMDSTR "netsh wlan show interface"
#endif

#if defined( HAVE_IWCONFIG ) && defined( CFG_PATH_IWCONFIG )
#define AP_WIPROG_CMDSTR CFG_PATH_IWCONFIG " 2>&1"
#endif

#define WIPROG_BUFSIZE 4096

#if defined( AP_WIPROG_CMDSTR )
static char wiprog_buffer[WIPROG_BUFSIZE];
#endif
/*
 * I couldn't find any official documentation on parsing the value under this
 * registry key but it seems that the SSID starting at byte 20 is pretty
 * consistent. This registry key can also be empty if the computer is not using
 * the default windows services for wireless networking. In addition, this
 * registry key seems to not be reliable i.e.  it may contain stale values and
 * not the current one the computer is connected to even if the key is called
 * "ActiveSettings".
 *
 * If we were on Vista, we could just fork a process and parse "netsh wlan show"
 *
 */
#define INTERFACE_REGKEY "SOFTWARE\\Microsoft\\WZCSVC\\Parameters\\Interfaces"
#define SUBKEYPATH_MAX_BUFLEN 256
#define SUBKEYVAL_MAX_BUFLEN 2048
#define SUBKEYVAL_INDEX 20
#define SUBKEY_MAX_ENTRIES 10 //so we don't have while( 1 ) when looping over subkeys

#define NETSH_WLAN_MAX_BUFLEN 8192

#if defined( AP_WIPROG_CMDSTR )
/*
 * Looks for the SSID in static global <wiprog_buffer>
 * 
 * Expects the following <marker><starter><nspaces>SSID_STRING_HERE<terminator>
 *
 */ 
static int get_ssidstring( uchar* buf, int blen, char* marker, char starter, int nspaces, char terminator )
{
    assert( buf != NULL && blen > 0 && marker != NULL );
    memset( buf, 0x00, blen );
    
    char* src = strstr( wiprog_buffer, marker );
    if( src == NULL ) {
        memset( wiprog_buffer, 0x00, sizeof( wiprog_buffer ) );
        memset( buf, 0x00, blen );
        return FAILURE;
    }
    
    while( ( src - wiprog_buffer ) < sizeof( wiprog_buffer ) && *src != starter ) {
        src++;
    }
    
    src += ( nspaces + 1 ); //seek to the ssid string itself

    uchar* dst = (uchar*)buf;
    while( ( src - wiprog_buffer ) < sizeof( wiprog_buffer ) && ( dst - buf ) < blen && *src != terminator ) {
        *dst = *src;
        src++;
        dst++;
    }
    
    if( dst != buf ) {
        //we could have copied to all of buf so we make sure it is null terminated 
        buf[blen - 1] = '\0'; 
        return SUCCESS;
    }
    else {
        memset( wiprog_buffer, 0x00, sizeof( wiprog_buffer ) );
        memset( buf, 0x00, blen );
        return FAILURE;
    }
}
#endif

#if defined( __APPLE__ )
static int get_ssid_airport( uchar* buf, int blen )
{
    assert( buf != NULL && blen > 0 );
    
    memset( wiprog_buffer, 0x00, sizeof( wiprog_buffer ) );
    if( exec_command( AP_WIPROG_CMDSTR, wiprog_buffer, sizeof( wiprog_buffer ) ) == FAILURE ) {
        memset( wiprog_buffer, 0x00, sizeof( wiprog_buffer ) );
        memset( buf, 0x00, blen );
        return FAILURE;
    }   
    
    return get_ssidstring( buf, blen, " SSID", ':', 1, '\n' );
}
#endif

#if defined( HAVE_IWCONFIG ) && defined( CFG_PATH_IWCONFIG )
static int get_ssid_iwconfig( uchar* buf, int blen )
{
    assert( buf != NULL && blen > 0 );
    
    memset( wiprog_buffer, 0x00, sizeof( wiprog_buffer ) );
    if( exec_command( AP_WIPROG_CMDSTR, wiprog_buffer, sizeof( wiprog_buffer ) ) == FAILURE ) {
        memset( wiprog_buffer, 0x00, sizeof( wiprog_buffer ) );
        memset( buf, 0x00, blen );
        return FAILURE;
    }   
    
    return get_ssidstring( buf, blen, " ESSID:", '"', 0, '"' );

}
#endif

#if defined( __CYGWIN__ )
static int get_ssid_regkey( uchar* buf, int blen )
{
    assert( buf != NULL && blen > 0 );
    
    OSVERSIONINFO winver;
    memset( &winver, 0x00, sizeof( OSVERSIONINFO ) );
    winver.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
        
    if( GetVersionEx( &winver ) == 0 ) {
        error( "could not get windows version, defaulting to XP" );
        winver.dwMajorVersion = 5; // we default to Windows XP
    }
    
    if( winver.dwMajorVersion == 6 ) { //Windows Vista
        memset( wiprog_buffer, 0x00, sizeof( wiprog_buffer ) );

        if( exec_command( AP_WIPROG_CMDSTR, wiprog_buffer, sizeof( wiprog_buffer ) ) == FAILURE ) {
            error( "could not execute netsh wlan command" );
            memset( buf, 0x00, sizeof( blen ) );
            return FAILURE;
        }
        
        return get_ssidstring( buf, blen, "SSID", ':', 1, '\r' ); 
    
    }
    else {
        char subkeyname[SUBKEYPATH_MAX_BUFLEN];
        char subkeyval[SUBKEYVAL_MAX_BUFLEN];
        DWORD sknlen = sizeof( subkeyname ); 
        DWORD skvlen = sizeof( subkeyval ); 
        memset( subkeyname, 0x00, sknlen );
        memset( subkeyval, 0x00, skvlen );
         
        DWORD rv;
        HKEY key;
        
        rv = RegOpenKeyEx( HKEY_LOCAL_MACHINE, INTERFACE_REGKEY, 0, KEY_READ, &key );
        if( rv != ERROR_SUCCESS || key == NULL ) {
            memset( buf, 0x00, blen );
            error( "could not open interface registry key" );
            return FAILURE;
        }
        
        int subkeyi = 0;
        int ssidfound = FALSE;
        char* subkeypath = NULL;
        HKEY subkey;
        
        while( subkeyi < SUBKEY_MAX_ENTRIES ) {
            rv = RegEnumKey( key, subkeyi, subkeyname, sknlen );
            if( rv != ERROR_SUCCESS ) {
                break;
            }
            
            subkeypath = concatstr( INTERFACE_REGKEY "\\", subkeyname );
            if( subkeypath == NULL ) {
                error( "could not build interface registry subkey" );       
                break;
            }
            
            rv = RegOpenKeyEx( HKEY_LOCAL_MACHINE, subkeypath, 0, KEY_READ, &subkey );
            if( rv != ERROR_SUCCESS || subkey == NULL ) {
                error( "could not open interface registry subkey" );
                break;
            }
            
            DWORD ktype = REG_BINARY;
            rv = RegQueryValueEx( subkey, "ActiveSettings", NULL, &ktype, (LPBYTE)&subkeyval, &skvlen );
            if( rv != ERROR_SUCCESS ) {
                error( "could not get value of interface registry subkey" );
                break;
            }
            
            char* ssidstr = &subkeyval[SUBKEYVAL_INDEX];
            int sslen = strlen( ssidstr );
            if( sslen > 0 &&  sslen < blen ) {
                memcpy( buf, ssidstr, sslen );
                buf[blen - 1] = '\0'; //make sure the string is NULL terminated
                ssidfound = TRUE;
                break;
            }
            // else we continue and look at other subkeys
            
            if( subkeypath != NULL ) { 
                free( subkeypath );
                subkeypath = NULL;
            }
            
            if( subkey != NULL ) { 
                RegCloseKey( subkey );
                subkey = NULL;
            }
            
            subkeyi++;
        }
        
        // clean up
        if( subkeypath != NULL ) { 
            free( subkeypath );
            subkeypath = NULL;
        }
        
        if( subkey != NULL ) { 
            RegCloseKey( subkey );
            subkey = NULL;
        }

        if( key != NULL ) { 
            RegCloseKey( key );
            key = NULL;
        }

        if( ssidfound == TRUE ) {
            return SUCCESS;
        }
        else {
            memset( buf, 0x00, blen );
            error( "could not find ssid" );
            return FAILURE;
        }
    }   
}
#endif

/*
 * Get the SSID of the AP we are connected to.
 *
 */ 
int ap_ssid_get( apdata_t* apd OUT )
{
    if( apd != NULL ) {
        memset( &apd->ssid, 0x00, AP_SSID_MAXLEN );

#if defined( __APPLE__ )
        return get_ssid_airport( apd->ssid, AP_SSID_MAXLEN );
#elif defined( __CYGWIN__ )
        return get_ssid_regkey( apd->ssid, AP_SSID_MAXLEN );
#elif defined( HAVE_IWCONFIG ) && defined( CFG_PATH_IWCONFIG )
        return get_ssid_iwconfig( apd->ssid, AP_SSID_MAXLEN );
#else
        return FAILURE;
#endif
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

