/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#if !defined( __CYGWIN__ )
    #include <ifaddrs.h>
#endif

#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>

#include "config.h"
#include "ipaddress.h"
#include "log.h"
#include "util.h"
#include "options.h"


#if defined( __CYGWIN__ )
    #include <windows.h>
#endif

#define EXTIP_MAX_BUFLEN 512
//#define EXTIP_PROVIDER_HOSTNAME "checkip.dyndns.org"
#define EXTIP_PROVIDER_PORT 80
#if 0
#define EXTIP_HTTPGET_STRING \
    "GET / HTTP/1.1\n" \
    "Host: " EXTIP_PROVIDER_HOSTNAME "\n"\
    "Connection: close\n\n"
#endif
#define EXTIP_GETSTR_BUFLEN 128

#define IPCONFIG_MAX_BUFLEN 8192

/*
 * Simple HTTP GET-ter
 *
 */ 
static int extip_httpget( char* servername, char* path, char* buffer, int blen )
{
    assert( buffer != NULL && blen > 0 );
    
    struct hostent* he = gethostbyname( servername );

    if( he != NULL ) {
        struct sockaddr_in epaddr;
        memset( &epaddr, 0x00, sizeof( epaddr ) );
        
        epaddr.sin_family = AF_INET;
        epaddr.sin_port = htons( EXTIP_PROVIDER_PORT );
        epaddr.sin_addr = *(struct in_addr*)he->h_addr_list[0];
        
        int epsock;
        if( ( epsock = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 ) {
            error( "could not create socket" );
            return FAILURE;
        }
        
        char get_request[EXTIP_GETSTR_BUFLEN];
        memset( get_request, 0x00, sizeof( get_request ) );
        snprintf( get_request, sizeof( get_request), 
                "GET /%s HTTP/1.1\nUser-Agent: Adeona\nHost: %s\nConnection: close\n\n", path, servername ); //path and servername should already have been "checked"

        if( connect( epsock, (struct sockaddr*)&epaddr, sizeof( epaddr ) ) == 0 ) {
            
            if( send( epsock, get_request, strlen( get_request ), 0 ) < 0 ) {
                error( "could not HTTP GET" );
                close( epsock );
                return FAILURE;
            }
            
            int nread = 0;
            int nbread = blen;
            char* bread = buffer;
            while( ( nread = recv( epsock, bread, nbread, 0 ) ) > 0 ) {
                bread += nread;
                nbread -= nread;
                if( nbread <= 0 ) {
                    break;
                }
            }

            if( nread < 0 ) {
                error( "could not complete HTTP GET" );
                memset( buffer, 0x00, blen );
                close( epsock );
                return FAILURE;
            }
            
            close( epsock );
            return SUCCESS;
        }
        else {
            error( "could not connect to external ip server" );
            return FAILURE;
        }
           
    }
    else {
        error( "external ip server ip lookup failed" );
        return FAILURE;   
    }
}

/*
 * Looks for the IP address string in <buffer>.
 *
 * Expects the IP address string to be <marker><starter><nspaces>IPADDRESS_HEREt<terminator> 
 * 
 * Note: static buffer for return value
 *
 */
static char* get_ipstring( char* buffer, int blen, char* marker, char starter, int nspaces, char terminator )
{
    assert( buffer != NULL && blen > 0 );
    
    static char parsedip[16];
    memset( parsedip, 0x00, sizeof( parsedip ) );

    char* src = strstr( buffer, marker );
    if( src == NULL ) {
        error( "could not find IP Address marker in buffer" );
        return NULL;
    }
    
    while( ( src - buffer ) < blen && *src != starter ) {
        src++;
    }
    
    src += ( nspaces + 1 ); //seek to IP string itself
    
    char* dst = parsedip;
    while( ( src - buffer ) < blen && ( dst - parsedip ) < sizeof( parsedip ) && *src != terminator  ) {
        *dst = *src;
        src++;
        dst++;
    }
    
    if( allowed_chars_only( parsedip, sizeof( parsedip ), "1234567890." ) == FALSE  ) {
        error( "parsed IP address contained invalid characters" );
        memset( parsedip, 0x00, sizeof( parsedip ) );
        return NULL;
    }
    
    return parsedip;
}

int get_externalip( struct in_addr* extip OUT )
{
    if( extip != NULL  && STROPT_ISLOADED( opts.extip_url ) ) {
        char ipbuf[EXTIP_MAX_BUFLEN];
        memset( ipbuf, 0x00, sizeof( ipbuf ) );
        
        char* path = memchr( opts.extip_url, '/', MAX_STROPTION_LEN );
        if( path == NULL ) {
            error( "could not find path in external ip url" );
            return FAILURE;
        }

        char servername[MAX_STROPTION_LEN];
        memset( servername, 0x00, MAX_STROPTION_LEN );
        memcpy( servername, opts.extip_url, path - opts.extip_url ); 
        
        path++;
        if( extip_httpget( servername, path, ipbuf, sizeof( ipbuf ) ) == FAILURE ) {
            error( "could not http get" );
            memset( extip, 0x00, sizeof( struct in_addr ) );
            return FAILURE;
        }
        
        char* ipstr = get_ipstring( ipbuf, sizeof(ipbuf ), "IP Address", ':', 1, '<' );
        if( ipstr != NULL ) {   
            if( inet_aton( ipstr, extip ) == 0 ) {
                error( "invalid ip address" );
                memset( extip, 0x00, sizeof( struct in_addr ) );
                return FAILURE;
            }

            return SUCCESS;
        }
        else {
            memset( extip, 0x00, sizeof( struct in_addr ) );
            return FAILURE;
        }
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}


int get_internalip( struct in_addr* intip OUT )
{
#if defined( __CYGWIN__ )
    //TODO: getifaddrs/freeifaddrs looks like will be available in Cygwin 1.7
    if( intip != NULL ) {
        OSVERSIONINFO winver;
        memset( &winver, 0x00, sizeof( OSVERSIONINFO ) );
        winver.dwOSVersionInfoSize = sizeof( OSVERSIONINFO );
        
        if( GetVersionEx( &winver ) == 0 ) {
            error( "could not get windows version, defaulting to XP" );
            winver.dwMajorVersion = 5; // we default to Windows XP
        }
        
        if( winver.dwMajorVersion == 6 ) { //Windows Vista
            //SIOCGIFCONF doesn't seem to give back wireless interfaces on Windows Vista.
            //We just look at the ipconfig output.
            
            char ipconfig[IPCONFIG_MAX_BUFLEN];
            memset( ipconfig, 0x00, sizeof( ipconfig ) );

            if( exec_command( "ipconfig", ipconfig, sizeof( ipconfig ) ) == FAILURE ) {
                error( "could not execute ipconfig command" );
                memset( intip, 0x00, sizeof( struct in_addr ) );
                return FAILURE;
            }
            
            char* ipstr = get_ipstring( ipconfig, sizeof( ipconfig ), "IPv4 Address", ':', 1, '\r' );
            if( ipstr != NULL ) {   
                if( inet_aton( ipstr, intip ) == 0 ) {
                    error( "invalid ip address" );
                    memset( intip, 0x00, sizeof( struct in_addr ) );
                    return FAILURE;
                }

                return SUCCESS;
            }
            else {
                memset( intip, 0x00, sizeof( struct in_addr ) );
                return FAILURE;
            }
            
        }
        else {
            struct ifconf ifc; 
            struct ifreq  *ifr; 
            struct ifreq  *ifend; 
            struct ifreq ifs[64];
            int sock; 
            
            if( ( sock = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 ) {
                error( "could not create socket" );
                return FAILURE;
            }
            
            ifc.ifc_len = sizeof( ifs );
            ifc.ifc_req = ifs;
            if( ioctl( sock,  SIOCGIFCONF, (char*)&ifc ) < 0 ) { 
                error( "could not get SIOCGIFCONF query" );
                close( sock );
                return FAILURE;
            }
            
            int found = 0;
            ifend = ifs + ( ifc.ifc_len / sizeof( struct ifreq ) );
            for( ifr=ifc.ifc_req; ifr<ifend; ifr++ ) {
                if( ifr->ifr_addr.sa_family == AF_INET ) {
                    if(ioctl( sock, SIOCGIFFLAGS, (char*)ifr ) < 0 ) {
                        continue;                
                    }
                    
                    if( ( ifr->ifr_flags & IFF_RUNNING ) && !( ifr->ifr_flags & IFF_LOOPBACK ) ) {
                        *intip = ( (struct sockaddr_in*)&ifr->ifr_addr )->sin_addr; 
                        if( intip->s_addr != 0 ) {
                            found = 1;
                            break;
                        }
                    }
                }
            }
            
            close( sock );
            
            if( !found ) {
                memset( intip, 0x00, sizeof( struct in_addr ) );
                return FAILURE;        
            }
        
            return SUCCESS;
        }
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }

#else   //get_internalip for Linux and Apple
    
    if( intip != NULL ) {
        struct ifaddrs* interfaces;

        if( getifaddrs( &interfaces ) < 0 ) {
            error( "could not get interfaces");
            memset( intip, 0x00, sizeof( struct in_addr ) );
            return FAILURE;
        }
        
        /* find the first active AF_INET interface that is not the local loopback */
        struct ifaddrs* cur = interfaces;
        int found = 0;
        while( cur != NULL ) {
            if( cur->ifa_addr->sa_family == AF_INET && !( cur->ifa_flags & IFF_LOOPBACK ) ) {
                *intip = ( (struct sockaddr_in*)cur->ifa_addr )->sin_addr;
                found = 1;
                break; 
            }
 
            cur = cur->ifa_next;
        }
        
        freeifaddrs( interfaces );
        
        if( !found ) {
            memset( intip, 0x00, sizeof( struct in_addr ) );
            return FAILURE;
        }
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }

#endif
}

