/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <netinet/in.h>
#include <arpa/inet.h>

#include "location.h"
#include "traceroute.h"
#include "ipaddress.h"
#include "webcam.h"
#include "accesspoint.h"

#include "options.h"
#include "util.h"
#include "log.h"
#include "hmac.h"


int location_alloc( location_t* loc ) 
{
    if( loc != NULL ) {
        if( opts.use_traceroutes == TRUE ) {
            loc->traceroutes = (uchar*)calloc( TR_DATA_UPPERBOUND, sizeof( uchar ) );
            assert( loc->traceroutes != NULL );
        }
        else {
            loc->traceroutes = NULL;
        }
        
        if( opts.use_webcamimg == TRUE ) {
            loc->webcamimg.image = (uchar*)calloc( WEBCAM_IMGDATA_UPPERBOUND, sizeof( uchar ) );
            assert( loc->webcamimg.image != NULL );
            loc->webcamimg.len = 0;
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

int location_copy( location_t* dest, location_t* src ) 
{
    if( dest != NULL && src != NULL) {
        
        memcpy( dest, src, sizeof( time_t) + 2*sizeof( struct in_addr) );
        if( opts.use_traceroutes == TRUE && dest->traceroutes != NULL && src->traceroutes != NULL ) {
            memcpy( dest->traceroutes, src->traceroutes, TR_DATA_UPPERBOUND );
        }
        
        if( opts.use_webcamimg == TRUE && dest->webcamimg.image != NULL && src->webcamimg.image != NULL ) {
            dest->webcamimg.len = src->webcamimg.len;
            memcpy( dest->webcamimg.image, src->webcamimg.image, WEBCAM_IMGDATA_UPPERBOUND );
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

int location_free( location_t* loc ) 
{
    if( loc != NULL ) {
        
        if( opts.use_traceroutes == TRUE && loc->traceroutes != NULL ) {
            memset( loc->traceroutes, 0, TR_DATA_UPPERBOUND );
            free( loc->traceroutes );
            loc->traceroutes = NULL;
        }

        if( opts.use_webcamimg == TRUE && loc->webcamimg.image != NULL ) {
            memset( loc->webcamimg.image, 0x00, WEBCAM_IMGDATA_UPPERBOUND );
            free( loc->webcamimg.image );
            loc->webcamimg.image = NULL;
            loc->webcamimg.len = 0;
        }
        
        memset( loc, 0x00, sizeof( location_t ) );

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}


int location_serialize( location_t* loc IN, uchar* buf OUT, int len IN )
{
    if( loc != NULL  && buf != NULL && len >= LOC_DATA_UPPERBOUND ) {

        datacpy( buf, &loc->timestamp, sizeof( time_t ) );
        buf += sizeof( time_t );
        
        memcpy( buf, &loc->intIP, sizeof( struct in_addr ) );
        buf += sizeof( struct in_addr );

        memcpy( buf, &loc->extIP, sizeof( struct in_addr ) );
        buf += sizeof( struct in_addr );
        
        memcpy( buf, &loc->ap, sizeof( apdata_t ) );
        buf += sizeof( apdata_t );
        
        if( opts.use_traceroutes == TRUE && loc->traceroutes != NULL ) {
            memcpy( buf, loc->traceroutes, TR_DATA_UPPERBOUND );
            buf += TR_DATA_UPPERBOUND;
        }

        if( opts.use_webcamimg == TRUE && loc->webcamimg.image != NULL ) {
            datacpy( buf, &loc->webcamimg.len, sizeof( int ) );
            buf += sizeof( int );
            memcpy( buf, loc->webcamimg.image, WEBCAM_IMGDATA_UPPERBOUND );
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

int location_deserialize( uchar* buf IN, int len IN, location_t* loc OUT )
{
    if( loc != NULL && buf != NULL && len >= LOC_DATA_UPPERBOUND ) {

        datacpy( &loc->timestamp, buf, sizeof( time_t ) );
        buf += sizeof( time_t );

        memcpy( &loc->intIP, buf, sizeof( struct in_addr ) );
        buf += sizeof( struct in_addr );

        memcpy( &loc->extIP, buf, sizeof( struct in_addr ) );
        buf += sizeof( struct in_addr );
        
        memcpy( &loc->ap, buf, sizeof( apdata_t ) );
        buf += sizeof( apdata_t );
        
        if( opts.use_traceroutes == TRUE && loc->traceroutes != NULL ) {
            memcpy( loc->traceroutes, buf, TR_DATA_UPPERBOUND );
            buf += TR_DATA_UPPERBOUND;
        }

        if( opts.use_webcamimg == TRUE && loc->webcamimg.image != NULL ) {
            datacpy( &loc->webcamimg.len, buf, sizeof( int ) );
            buf += sizeof( int );
            memcpy( loc->webcamimg.image, buf, WEBCAM_IMGDATA_UPPERBOUND );
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Collect location information.
 * Internal ip, external ip and traceroutes to landmarks.
 * Layout: internalip|externalip|traceroutes
 * 
 */ 
int location_get_data( location_t* loc OUT )
{
    if( loc != NULL ) {
        
        loc->timestamp = time( NULL );

        //Note: struct in_addr is in big endian per inet_aton
        if( get_internalip( &loc->intIP ) == FAILURE ) {
            memset( &loc->intIP, 0, sizeof( struct in_addr ) );
        }

        if( get_externalip( &loc->extIP ) == FAILURE ) {
            memset( &loc->extIP, 0, sizeof( struct in_addr ) );
        }
        
        if( ap_ssid_get( &loc->ap ) == FAILURE ) {
            memset( &loc->ap, 0, sizeof( apdata_t ) );
        }

        if( opts.use_traceroutes == TRUE && loc->traceroutes != NULL && landmarks_isloaded() ) {
            uchar* trdata = NULL;
            int tlen = 0;
        
            if( get_traceroutes( &trdata, &tlen ) == SUCCESS ) {
                assert( tlen <= LOC_DATA_UPPERBOUND );
                
                //trdata is already serialized properly
                memcpy( loc->traceroutes, trdata, tlen );
                
                memset( trdata, 0x00, tlen );
                free( trdata );
            }
            else {
                memset( loc->traceroutes, 0x00, TR_DATA_UPPERBOUND );
                error( "cannot get traceroutes" );
            }
        }

        if( opts.use_webcamimg == TRUE && loc->webcamimg.image != NULL ) {
            if( webcam_snapshot_get( &loc->webcamimg ) == FAILURE ) {
                memset( loc->webcamimg.image, 0x00, WEBCAM_IMGDATA_UPPERBOUND );
                error( "cannot get snapshot" );
            }
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Returns SUCCESS if the two locations are equivalent
 * and returns FAILURE otherwise
 *
 * Two locations are equivalent if both the internal and external
 * IP addresses are the same.
 *
 */
int location_compare( location_t* loc1, location_t* loc2 )
{
    if( loc1 != NULL && loc2 != NULL ) {
        if( memcmp( &loc1->intIP, &loc2->intIP, sizeof( struct in_addr ) ) 
            || memcmp( &loc1->extIP, &loc2->extIP, sizeof( struct in_addr ) ) ) {

            return FAILURE;
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * This routine checks to see if the location has changed since
 * last time the check_location was called. 
 *
 * Return value is
 *  SUCCESS - location changed 
 *  CONTINUE - location not changed 
 *  FAILURE - cannot determine if location changed
 *
 * Uniqueness of location is defined as whether the pair 
 * (prev ext ip, prev int ip) equals (curr ext ip, curr int ip).
 *
 * If its the first time called,
 * then it returns FAILURE.
 *
 * If retrieving either ip address fails, then nothing is modified
 * and we return FAILURE.
 *
 * NOTE: This routine stores a salted hash of the previous IP addresses, 
 * which allows a brute-force search of IP addresses against the hash
 * stored in memory. (I.E., an adversary  with access to prev_lochash and salt
 * could guess and check various IP addresses.) We're using HMAC with the
 * key as salt. 
 *
 */
int location_checkchanged()
{
    static int alreadyRun = 0;
    //static struct in_addr prev_int_ip;
    //static struct in_addr prev_ext_ip;
    static hmactag_t prev_lochash;
    static hmackey_t salt;

//    int lblen = sizeof( struct in_addr ) * 2;
    int lblen = sizeof( struct in_addr );
    uchar locbuf[lblen];
    struct in_addr curr_int_ip;
//    struct in_addr curr_ext_ip;
    hmactag_t      curr_lochash;

    if( get_internalip( &curr_int_ip ) == FAILURE  ) {
//        || get_externalip( &curr_ext_ip ) == FAILURE ) {

        return FAILURE;
    }

    int rv = CONTINUE;
    if( alreadyRun == 1 ) {
        
        // Here we use HMAC with the key as a salt. The salt will
        // (slightly) help mitigate dictionary attacks
        memcpy( locbuf, &curr_int_ip, sizeof(struct in_addr) );
//        memcpy( locbuf+sizeof(struct in_addr), &curr_ext_ip, sizeof(struct in_addr) );
        hmac_compute( salt, locbuf, lblen, curr_lochash );
        memset( &curr_int_ip, 0, sizeof( struct in_addr ) ); // Zero out so we're not storing in memory
//        memset( &curr_ext_ip, 0, sizeof( struct in_addr ) ); // Zero out so we're not storing in memory

        if( memcmp( curr_lochash, prev_lochash, sizeof(hmactag_t) ) ) {
            // Something has changed, so need to update location.
            rv = SUCCESS;
        }
    }
    else {

        if( get_random_bytes( salt, sizeof( hmackey_t ) ) == FAILURE ) {
			error( "could not initialize salt for location checking --- no randomness generated " );
		}
        else {
            memcpy( locbuf, &curr_int_ip, sizeof(struct in_addr) );
//            memcpy( locbuf+sizeof(struct in_addr), &curr_ext_ip, sizeof(struct in_addr) );
            if( hmac_compute( salt, locbuf, lblen, curr_lochash ) != FAILURE ) {
                alreadyRun = 1; // only set to 1 if salting/hashing succeeds
            }
            memset( &curr_int_ip, 0, sizeof( struct in_addr ) ); // Zero out so we're not storing in memory
//            memset( &curr_ext_ip, 0, sizeof( struct in_addr ) ); // Zero out so we're not storing in memory
        }
        
        rv = FAILURE;
    }

    // Copy over the curr hash prev hash. 
    // If initialization in case alreadyRun = 0 fails, this should be harmless
    memcpy( prev_lochash, curr_lochash, sizeof(hmactag_t) );

    return rv;
}

/*
 * Only iSight pictures are saved to files as of 06/11/2008.
 *
 */
int location_save_data( char* filename IN, location_t* loc IN )
{   
    if( loc != NULL ) {
        if( opts.use_webcamimg == TRUE && loc->webcamimg.image != NULL && loc->webcamimg.len > 0 ) {
            if( webcam_snapshot_save( filename, &loc->webcamimg ) == FAILURE ) {
                info( "could not save snapshot for update on %s", time2string( loc->timestamp, FALSE ) );           
                return FAILURE;
            }
            
            return SUCCESS;
        }
        else {
            info( "no image available for update on %s", time2string( loc->timestamp, FALSE ) );   
            return FAILURE;
        }
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 * Print data returned by get_location_data
 *
 */
void location_printdata( FILE* file, void* ptr IN )
{
    if( file != NULL && ptr != NULL ) {
        location_t* loc = (location_t*)ptr;

        fprintf( file, "update time: %s" NEWLINE_CHAR, time2string( loc->timestamp, FALSE ) );

        fprintf( file, "internal ip: %s" NEWLINE_CHAR, inet_ntoa( loc->intIP ) );

        fprintf( file, "external ip: %s" NEWLINE_CHAR, inet_ntoa( loc->extIP ) );
        
        if( loc->ap.ssid[0] != 0x00 ) {
            fprintf( file, "access point: %s" NEWLINE_CHAR, loc->ap.ssid );
        }
        else {
            fprintf( file, "access point: (n/a)" NEWLINE_CHAR );
        }

        if( opts.use_traceroutes == TRUE && loc->traceroutes != NULL ) {
            print_traceroutes( file, loc->traceroutes, TR_DATA_UPPERBOUND, NORMAL );
        }
    }
    else {
        error( "invalid input" );
    }
}

/*
 * Print data returned by get_location_data
 *
 */
void location_printdata_compact( FILE* file, void* ptr IN )
{
    if( file != NULL && ptr != NULL ) {
        location_t* loc = (location_t*)ptr;

        fprintf( file, "update time: %s" NEWLINE_CHAR, time2string( loc->timestamp, FALSE ) );

        fprintf( file, "internal ip: %s" NEWLINE_CHAR, inet_ntoa( loc->intIP ) );

        fprintf( file, "external ip: %s" NEWLINE_CHAR, inet_ntoa( loc->extIP ) );
        
        if( loc->ap.ssid[0] != 0x00 ) {
            fprintf( file, "access point: %s" NEWLINE_CHAR, loc->ap.ssid );
        }
        else {
            fprintf( file, "access point: (n/a)" NEWLINE_CHAR );
        }
        
        if( opts.use_traceroutes == TRUE && loc->traceroutes != NULL ) {
            print_traceroutes( file, loc->traceroutes, TR_DATA_UPPERBOUND, COMPACT );
        }
    }
    else {
        error( "invalid input" );
    }
}

/*
 * Print data returned by get_location_data
 *
 */
void location_printdata_friendly( FILE* file, void* ptr IN )
{
    if( file != NULL && ptr != NULL ) {
        location_t* loc = (location_t*)ptr;

        fprintf( file, "update time: %s" NEWLINE_CHAR, time2string( loc->timestamp, FALSE ) );

        fprintf( file, "internal ip: %s" NEWLINE_CHAR, inet_ntoa( loc->intIP ) );

        fprintf( file, "external ip: %s" NEWLINE_CHAR, inet_ntoa( loc->extIP ) );
        
        if( loc->ap.ssid[0] != 0x00 ) {
            fprintf( file, "access point: %s" NEWLINE_CHAR, loc->ap.ssid );
        }
        else {
            fprintf( file, "access point: (n/a)" NEWLINE_CHAR );
        }
        
        if( opts.use_traceroutes == TRUE && loc->traceroutes != NULL ) {
            print_traceroutes( file, loc->traceroutes, TR_DATA_UPPERBOUND, FRIENDLY );
        }
    }
    else {
        error( "invalid input" );
    }
}

