/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <netinet/in.h>
#include <netdb.h>

#include "util.h"
#include "log.h"
#include "traceroute.h"
#include "landmarks.h"
#include "config.h"

#define TR_BUFLEN 4096

#if defined( __CYGWIN__ )
    #define TR_CMDSTR "pathping.exe -n -q 1 -h " TR_HOP_LIMIT_STR " "
#else
    #define TR_CMDSTR CFG_PATH_TRACEROUTE " 2>&1 -n -q 1 -m " TR_HOP_LIMIT_STR " "
#endif

#if 0
#define lastline( ptr ) ( *( ptr + 1 ) == '\0' ) 
#define nextline( ptr ) while( *( ptr++ ) != '\n' && *ptr != '\0' )
#define skiplines( n, ptr ) do { \
        int i; \
        for( i=0; i<n; i++ ) { \
            nextline( ptr ); \
        } \
    } while( 0 );
#endif

#define lastline( ptr ) ( ( ptr != NULL ) && ( *( ptr + 1 ) == '\0' ) )

#define nextline( buffer, blen, ptr ) while( ( ptr >= buffer ) \
        && ( ( ptr - buffer ) < blen )  \
        && ( ( ptr != NULL ) && ( *( ptr++ ) != '\n' ) && ( *ptr != '\0' ) ) )

#define skiplines( buffer, blen, n, ptr ) do { \
        int i; \
        for( i=0; i<n; i++ ) { \
            nextline( buffer, blen, ptr ); \
        } \
    } while( 0 );


/*
 * Fill in probe_t structure <pr_buffer> with traceroute information
 * from <tr_buffer>
 * 
 * Note: On Windows, we use pathping.exe and the resolution for the
 * RTT is only milliseconds.
 *
 */ 
static void get_probe_struct( struct in_addr target, char* tr_buffer, int trblen, probe_t* pr_buffer )
{
    assert( tr_buffer != NULL && pr_buffer != NULL );
    
    // set the target
    pr_buffer->target = target;
    
    // parse the traceroute output
#if defined( __CYGWIN__ )
    
    // for windows, we use pathping instead.
    char* read = tr_buffer;
    int maxhop = 0;

    skiplines( tr_buffer, trblen, 4, read );
    
    // parse the hops part
    while( ( read - tr_buffer ) < trblen && !lastline( read ) && ( *read != '\r' || *read != 'C' ) ) {
        int hopnum;
        char ipaddr[16]; //sscanf appends '\0'
        memset( ipaddr, 0x00, sizeof( ipaddr ) );
        
        if( sscanf( read, "%d %15s", &hopnum, ipaddr ) <= 0 ) {
            break;
        }
        
        assert( hopnum > 0 && hopnum <= TR_HOP_LIMIT && maxhop < hopnum );
        maxhop = hopnum;
        
        if( inet_aton( ipaddr, &pr_buffer->hops[hopnum - 1].ip ) == 0 ) {
            memset( &pr_buffer->hops[hopnum - 1].ip, 0x00, sizeof( struct in_addr ) );
        }
        
        nextline( tr_buffer, trblen, read );
    }
    
    // if the last hop was not reachable, pathping won't print an
    // extra line break before "Computing statistics"
    if( ( read - tr_buffer ) < trblen && *read == 'C' ) { 
        skiplines( tr_buffer, trblen, 5, read );
    }
    else {
        skiplines( tr_buffer, trblen, 6, read );
    }

    // now parse the RTTs part...the 0th hop has already been skipped    
    while( ( read - tr_buffer ) < trblen && !lastline( read ) && *read != '\r' ) {
        int hopnum;
        float rtt = 0.0f;
        if( sscanf( read, "%d %fms", &hopnum, &rtt ) <= 0 ) {
            break;
        }
        
        //for some reason, pathping will give back more than "-h TR_HOP_LIMIT" sometimes
        if( hopnum > maxhop ) {
            break;
        }
        
        assert( hopnum > 0 && hopnum <= TR_HOP_LIMIT );

        pr_buffer->hops[hopnum - 1].rtt = rtt;
        
        skiplines( tr_buffer, trblen, 2, read );
    }
    
#else
    
    char* read = tr_buffer;
    nextline( tr_buffer, trblen, read );

    while( ( read - tr_buffer ) < trblen && !lastline( read ) ) {
        int hopnum;
        char ipaddr[16]; //sscanf appends '\0'
        float rtt;
        
        if( sscanf( read, "%d %15s %f", &hopnum, ipaddr, &rtt ) <= 0 ) {
            break;
        }
        
        assert( hopnum <= TR_HOP_LIMIT );

        if( ipaddr[0] == '*' ) {
            memset( &pr_buffer->hops[hopnum - 1], 0x00, sizeof( hop_t ) );
        }
        else {
            if( inet_aton( ipaddr, &pr_buffer->hops[hopnum - 1].ip ) == 0 ) {
                memset( &pr_buffer->hops[hopnum - 1].ip, 0x00, sizeof( struct in_addr ) );
            }
            pr_buffer->hops[hopnum - 1].rtt = rtt;
        }

        nextline( tr_buffer, trblen, read );
    }
    
#endif
}

static probe_t* do_traceroute( struct in_addr target )
{
    // static buffers for processing traceroute data
    static char tr_buffer[TR_BUFLEN];
    static probe_t pr_buffer;
    memset( tr_buffer, 0x00, sizeof( tr_buffer ) );
    memset( &pr_buffer, 0x00, sizeof( pr_buffer ) );
    
    char* cmd = concatstr( TR_CMDSTR, inet_ntoa( target ) );
    if( exec_command( cmd, tr_buffer, sizeof( tr_buffer ) ) == FAILURE ) {
        memset( tr_buffer, 0x00, sizeof( tr_buffer ) );
        free( cmd );
        cmd = NULL;
        return NULL;
    }   
    free( cmd );
    cmd = NULL;
    
    get_probe_struct( target, tr_buffer, TR_BUFLEN, &pr_buffer );
    
    return ( &pr_buffer );
}

static void serialize_probe( probe_t* src, uchar* buf )
{
    assert( src != NULL && buf != NULL );
    
    memcpy( buf, &src->target, sizeof( struct in_addr ) );
    buf += sizeof( struct in_addr );

    int i;
    for( i=0; i<TR_HOP_LIMIT; i++ ) {
        memcpy( buf, &src->hops[i].ip, sizeof( struct in_addr ) );
        buf += sizeof( struct in_addr );
        
        datacpy( buf, &src->hops[i].rtt, sizeof( float ) );
        buf += sizeof( float );
    }
}

static void deserialize_probe( probe_t* src, uchar* buf )
{
    assert( src != NULL && buf != NULL );
    
    memcpy( &src->target, buf, sizeof( struct in_addr ) );
    buf += sizeof( struct in_addr );

    int i;
    for( i=0; i<TR_HOP_LIMIT; i++ ) {
        memcpy( &src->hops[i].ip, buf, sizeof( struct in_addr ) );
        buf += sizeof( struct in_addr );
        
        datacpy( &src->hops[i].rtt, buf, sizeof( float ) );
        buf += sizeof( float );
    }
}

static char* gethostbyaddr_wrap( struct in_addr ip )
{
    static char hostname[128];
    memset( hostname, 0x00, sizeof( hostname ) );

    struct hostent* he = gethostbyaddr( (char*)&ip, sizeof( struct in_addr ), AF_INET );
    if( he != NULL ) {
        snprintf( hostname, sizeof( hostname ), "%s", he->h_name );    
    }
    else {
        snprintf( hostname, sizeof( hostname ), "could not resolve" );    
    }

    return hostname;
}

static void tr_print_normal( FILE* file, uchar* data, int len )
{
    assert( file != NULL && data != NULL && len > 0 );

    probe_t pr;
    memset( &pr, 0x00, sizeof( probe_t ) );
    
    int i, j;
    for( i=0; i<TR_NPROBES; i++ ) {
        deserialize_probe( &pr, data );
        fprintf( file, "target: %s" NEWLINE_CHAR, inet_ntoa( pr.target ) );
        for( j=0; j<TR_HOP_LIMIT; j++ ) {
            if( pr.hops[j].rtt == 0.0f ) {
                fprintf( file, "%d * 0.000" NEWLINE_CHAR, j+1 ); 
            }
            else {
                fprintf( file, "%d %s %.3f" NEWLINE_CHAR, j+1, inet_ntoa( pr.hops[j].ip ), pr.hops[j].rtt ); 
            }
        }
        data += sizeof( probe_t );
    }
}

static void tr_print_compact( FILE* file, uchar* data, int len )
{
    assert( file != NULL && data != NULL && len > 0 );
    
    probe_t pr;
    memset( &pr, 0x00, sizeof( probe_t ) );
    
    int i, j;
    for( i=0; i<TR_NPROBES; i++ ) {
        deserialize_probe( &pr, data );
        fprintf( file, "target: %s ", inet_ntoa( pr.target ) );
        for( j=0; j<TR_HOP_LIMIT; j++ ) {
            if( pr.hops[j].rtt == 0.0f ) {
                fprintf( file, "[%d] * 0.000 ", j+1 ); 
            }
            else {
                fprintf( file, "[%d] %s %.3f ", j+1, inet_ntoa( pr.hops[j].ip ), pr.hops[j].rtt ); 
            }
        }
        fprintf( file, NEWLINE_CHAR );
        data += sizeof( probe_t );
    }

}

static void tr_print_friendly( FILE* file, uchar* data, int len )
{
    assert( file != NULL && data != NULL && len > 0 );
    
    probe_t pr;
    memset( &pr, 0x00, sizeof( probe_t ) );
    
    int i, j;
    for( i=0; i<TR_NPROBES; i++ ) {
        deserialize_probe( &pr, data );

        //fprintf( file, "Traceroute target was: %s (%s)" NEWLINE_CHAR, inet_ntoa( pr.target ), gethostbyaddr_wrap( pr.target ) );
        fprintf( file, "Nearby routers:" NEWLINE_CHAR );
        int k = 0;
        for( j=0; j<TR_HOP_LIMIT; j++ ) {
            if( pr.hops[j].ip.s_addr != 0 && memcmp( &pr.hops[j].ip, &pr.target, sizeof( struct in_addr ) ) != 0 ) {
                k++;
                fprintf( file, "%d %.3fms %s (%s)" NEWLINE_CHAR, k, pr.hops[j].rtt, inet_ntoa( pr.hops[j].ip ), gethostbyaddr_wrap( pr.hops[j].ip ) ); 
            }
        }
        
        if( k == 0 ) {
            fprintf( file, "no routers found" NEWLINE_CHAR );
        }

        data += sizeof( probe_t );
    }
}

/*
 * Caller must deallocate <data>. Currently, we know the size but
 * we'll do it like this anyway in case we want to change it (again).
 *
 */
int get_traceroutes( uchar** data OUT, int* len OUT )
{
    if( landmarks_isloaded() == FALSE ) {
        error( "no landmarks available" );
        *data = NULL;
        *len = 0;
        
        return FAILURE;
    }
    
    if( data != NULL && len != NULL ) {
        int dlen = TR_DATA_UPPERBOUND;
        uchar* allp = (uchar*)malloc( dlen );
        assert( allp != NULL );
        
        uchar* cur = allp;
        int i;
        for( i=0; i<TR_NPROBES; i++ ) {
            probe_t* onep = do_traceroute( landmarks_get() );
            if( onep != NULL ) {
                serialize_probe( onep, cur );
            }
            else {
                memset( cur, 0x00, sizeof( probe_t ) );
            }

            cur += sizeof( probe_t );
        }
        assert( cur - allp == dlen );
        
        *len = dlen;
        *data = allp;

        return SUCCESS;

    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

int print_traceroutes( FILE* file, uchar* data IN, int len IN, trpstyle_t style IN )
{
    if( data != NULL && len > 0 ) {
        
        switch( style ) {
            case COMPACT :
                tr_print_compact( file, data, len );
                break;
            case FRIENDLY :
                tr_print_friendly( file, data, len );
                break;
            case NORMAL : /* fall through */ 
            default: 
                tr_print_normal( file, data, len );
                break;
        }

        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

