/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <unistd.h>
#include <net/if.h>
#include <sys/ioctl.h>

#include "gateways.h"
#include "options.h"
#include "log.h"
#include "util.h"

static char** _gateways = NULL;
static int _num_gateways = 0;

#define GW_MAX_NAMELEN 1024
#define GW_FSCANF_FMT "%1024s"

#define GW_HTTPGET_PORT 80
#define GW_GETSTR_BUFLEN 128
#define GW_RECV_BUFLEN 1024
#define HTTP_HEADER_END_CRLF "\r\n\r\n"
#define HTTP_HEADER_END_LF "\n\n"

/*
 *
 * Caller must deallocate <update>
 *
 *
 */ 
static int get_update( char* servername, char* path, char** update, int* uplen )
{
    struct hostent* he = gethostbyname( servername );
    if( he != NULL ) {
        struct sockaddr_in gwaddr;
        memset( &gwaddr, 0x00, sizeof( gwaddr ) );
        
        gwaddr.sin_family = AF_INET;
        gwaddr.sin_port = htons( GW_HTTPGET_PORT );
        gwaddr.sin_addr = *(struct in_addr*)he->h_addr_list[0];
        
        int gwsock;
        if( ( gwsock = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 ) {
            error( "could not create socket" );
            return FAILURE;
        }
        
        char get_request[GW_GETSTR_BUFLEN];
        memset( get_request, 0x00, sizeof( get_request ) );
        snprintf( get_request, sizeof( get_request), 
                "GET /%s HTTP/1.1\nUser-Agent: Adeona\nHost: %s\nConnection: close\n\n", path, servername ); //path and servername should already have been "checked"
        
        if( connect( gwsock, (struct sockaddr*)&gwaddr, sizeof( gwaddr ) ) == 0 ) {
            if( send( gwsock, get_request, strlen( get_request ), 0 ) < 0 ) {
                error( "could not HTTP GET" );
                close( gwsock );
                return FAILURE;
            }
            
            char buf[GW_RECV_BUFLEN];
            memset( buf, 0x00, sizeof( buf ) );
            int nread = 0;
            while( ( nread = recv( gwsock, buf, sizeof( buf ), 0 ) ) > 0 ) {
                *update = (char*)realloc( *update, *uplen + nread );
                memcpy( *update + *uplen, buf, nread );
                *uplen += nread;   
            }

            if( nread < 0 ) {
                warn( "could not complete HTTP GET" );
            }
            
            close( gwsock );
            return SUCCESS;
        }
        else {
            error( "could not connect to external ip server" );
            return FAILURE;
        }
    }
    else {
        error( "external ip server ip lookup failed" );
        return FAILURE;
    }
    
}

int gateways_load( char* filename IN )
{
    if( filename != NULL ) {
        FILE* gwfile = fopen( filename, "r" );
        if( gwfile == NULL ) {
            error( "could not open gateways file" );
            _gateways = NULL;
            _num_gateways = 0;
            return FAILURE;
        }
        
        assert( _gateways == NULL && _num_gateways == 0 );

        char hostname[GW_MAX_NAMELEN];
        memset( hostname, 0x00, GW_MAX_NAMELEN );
        while( fscanf( gwfile, GW_FSCANF_FMT, hostname ) != EOF ) {
            
            if( allowed_chars_only( hostname, GW_MAX_NAMELEN, HOSTNAME_ALLOWED_CHARS ) == FALSE ) {
                continue;
            }
            
            _num_gateways++;
            _gateways = (char**)realloc( _gateways, _num_gateways * sizeof( char* ) );
            assert( _gateways != NULL );
            char* newgw = NULL;
            int ngwlen = strlen( hostname ) + 1; //hostname is NULL terminated by fscanf, safe to use strlen;
            newgw = (char*)malloc( ngwlen ); 
            assert( newgw  != NULL );
                
            memcpy( newgw, hostname, ngwlen ); //we copy the null terminator as well
            newgw[ngwlen - 1] = '\0'; // paranoid

            _gateways[_num_gateways - 1] = newgw;
        }

        fclose( gwfile );
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

const char* gateways_get_random( void )
{   
    if( _gateways != NULL && _num_gateways > 0 ) {
        return _gateways[get_random_uint( _num_gateways )];
    }
    else {
        return NULL;   
    }
}

const char* gateways_get( int index )
{   
    if( _gateways != NULL && _num_gateways > 0 ) {
        if( index >= 0 && index < _num_gateways  ) {
            return _gateways[index];
        }
        else {
            return NULL;
        }
    }
    else {
        return NULL;   
    }
}

int gateways_isloaded( void )
{
    return ( ( _gateways != NULL && _num_gateways > 0 ) ? TRUE:FALSE );
}

/*
 * url is expected to be updater.hostname.com/path/to/gateways_update_file
 *
 */ 
int gateways_update_get( char* url IN )
{
    if( url != NULL ) {
        char* update = NULL;
        int uplen = 0;

        char* path = memchr( url, '/', MAX_STROPTION_LEN );
        if( path == NULL ) {
            error( "could not find path in gateways update url" );
            return FAILURE;
        }
        
        char servername[MAX_STROPTION_LEN];
        memset( servername, 0x00, MAX_STROPTION_LEN );
        memcpy( servername, url, path - url );
        path++;

        get_update( servername, path, &update, &uplen );
        
        if( update != NULL ) {
            if( uplen <= 0 ) {
                error( "no update data" );
                free( update );
                return FAILURE;
            }
            
            char* gateways = strstr( update, HTTP_HEADER_END_CRLF );
            if( gateways == NULL ) {
                gateways = strstr( update, HTTP_HEADER_END_LF );
                if( gateways == NULL ) {
                    error( "could not find end of http header" );
                    free( update );
                    return FAILURE;
                }
            }
            
            //skip the HTTP header to the start of gateways data
            if( *gateways == '\r' ) {
                gateways += strlen( HTTP_HEADER_END_CRLF );
            }
            else {
                gateways += strlen( HTTP_HEADER_END_LF );
            }
            
            if( ( gateways - update ) >= uplen ) {
                error( "update has HTTP header but empty" );
                free( update );
                return FAILURE;
            }

            //TODO: A simple check against webpage redirection e.g. wifi
            //hotspots redirect you to a login page or a "terms of use" page
            //before you can use it.
            if( *gateways == '<'  || strstr( gateways, "<html>" ) != NULL || strstr( gateways, "<body>" ) != NULL ) {
                error( "not a gateway file" );
                free( update );
                return FAILURE;
            }

            FILE* gwfile = fopen( opts.gateways_filename, "w" ); 
            if( gwfile == NULL ) {
                error( "could not open gateway file for updating" );
                free( update );
                return FAILURE;
            }
        
            int realuplen = uplen - ( gateways - update );
            int gwrite = fwrite( gateways, sizeof( char ), realuplen, gwfile );
            if( gwrite < realuplen ) {
                error( "not able to save entire update" );
                free( update );
                fclose( gwfile );
                return FAILURE;;
            }

            free( update );
            fclose( gwfile );
            return SUCCESS;

        }
        else {
            error( "could not get updated gateways" );   
            return FAILURE;
        }
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
    
}

void gateways_free( void )
{
    if( _gateways != NULL ) {
        int i;
        for( i=0; i<_num_gateways; i++ ) {
            if( _gateways[i] != NULL ) {
                free( _gateways[i] );
                _gateways[i] = NULL;
            }
        }
        free( _gateways );

        _gateways = NULL;
        _num_gateways = 0;
    }
    else {
        _gateways = NULL;
        _num_gateways = 0;
    }
}
