/*
 * Copyright (c) 2007-2008 University of Washington
 * 
 * All rights reserved. For information, contact the Adeona research
 * team at adeona@cs.washington.edu
 *
 * See the file License.txt included in this distribution for details.
 *
 */
#include <netdb.h>
#include <arpa/inet.h>

#include "remstore.h"
#include "options.h"
#include "gateways.h"
#include "log.h"
#include "util.h"

#define RS_MAX_TRYSERVER 2
#define RS_MAX_TRYSERVERCACHE 10

#define RS_MAX_PUTTRY 4
#define RS_MAX_GETTRY 4

/*
 * Retrieves a list of gateways specified by GWUPDATE_URL if USE_GWUPDATE is
 * set. 
 *
 */ 
static void update_gateways_list( void )
{
    if( opts.use_gwupdate && STROPT_ISLOADED( opts.gwupdate_url ) ) {
        if( gateways_update_get( opts.gwupdate_url ) == SUCCESS ) {
            debug( "successfully updated gateways file" );
            
            gateways_free();
            
            if( STROPT_ISLOADED( opts.gateways_filename ) && gateways_load( opts.gateways_filename ) == SUCCESS ) {
                debug( "successfully loaded new gateways file" );
            }
            else {
                error( "could not load updated gateways" );
            }
        }
        else {
            error( "could not get gateways update" );
        }
    }
}

/* used to keep track of chunks when reassembling */
typedef struct _chunktrak_t
{
    uchar* data;
    int dlen;
} chunktrak_t;

static void serialize_header( chunkhdr_t* header, uchar* buffer )
{
    uchar* cur = buffer;
    
    memcpy( cur, header->ident, RS_CHUNKHDR_MAGIC_LEN );
    cur += RS_CHUNKHDR_MAGIC_LEN;
    
    datacpy( cur, &header->numchunks, sizeof( uint ) );
    cur += sizeof( uint );
    datacpy( cur, &header->currchunk, sizeof( uint ) );
    cur += sizeof( uint );
    datacpy( cur, &header->chunklen, sizeof( uint ) );
    cur += sizeof( uint );
    assert( cur - buffer == sizeof( chunkhdr_t ) );
}

static void deserialize_header( chunkhdr_t* header, uchar* buffer )
{
    uchar* cur = buffer;
    
    memcpy( header->ident, cur, RS_CHUNKHDR_MAGIC_LEN );
    cur += RS_CHUNKHDR_MAGIC_LEN;
    
    datacpy( &header->numchunks, cur, sizeof( uint ) );
    cur += sizeof( uint );
    datacpy( &header->currchunk, cur, sizeof( uint ) );
    cur += sizeof( uint );
    datacpy( &header->chunklen, cur, sizeof( uint ) );
    cur += sizeof( uint );
    assert( cur - buffer == sizeof( chunkhdr_t ) );
}

/*
 * If the hostname is opendht.nyuld.net i.e. OASIS, this will
 * alternate between the two ips it seems to give in subsequent calls.
 * 
 * WARNING: gethostbyname return value is UNCHECKED
 *
 */
static struct in_addr find_ip( const char* hostname )
{
    static struct in_addr previp; 
    
    struct hostent* he = gethostbyname( hostname );
    if( he != NULL ) {
        int i;
        for( i=0; he->h_addr_list[i] != NULL; i++ ) {
            struct in_addr newip = *(struct in_addr*)he->h_addr_list[i];
            if( previp.s_addr != newip.s_addr ) {
                previp = newip;
                break;
            }
        }
    }
    else {
        previp.s_addr = 0;
    }
    
    return previp;
}

#if 0
/*
 * Simple backup plan is to try the gateways list
 * 
 * Should only be called when gateways_isloaded() is TRUE
 *
 */
static struct in_addr find_from_gateways( void  )
{
    struct in_addr ret;
    ret.s_addr = 0;
    
    if( gateways_isloaded() == TRUE ) {
        const char* svr = gateways_get();
        if( svr != NULL ) {
            struct hostent* he = gethostbyname( svr );
            if( he != NULL ) {
                ret = *(struct in_addr*)he->h_addr_list[0];
            }
        }
    }

    return ret;
}
#endif

/*
 * Finds a gateway to connect to.
 *
 * If <first> is 0, we try to get the IP of the first server in the list.
 *
 */
static struct in_addr find_from_gateways( int first  )
{
    struct in_addr ret;
    memset( &ret, 0x00, sizeof( struct in_addr ) );

    if( gateways_isloaded() == TRUE ) {
        const char* svr = NULL;
        if( first == 0 ) {
            svr = gateways_get( 0 );
        }
        else {
            svr = gateways_get_random();
        }

        if( svr != NULL ) {
            struct hostent* he = gethostbyname( svr );
            if( he != NULL ) {
                ret = *(struct in_addr*)he->h_addr_list[0];
            }
        }
    }
    
    return ret;
}

#if 0
/*
 * This looks confusing but <try> is the i-th reconnect we've done
 * <try> allows switching between trying to connect to either the
 * supplied server or one of the servers in the gateways file
 *
 */ 
static int try_connect( int* try )
{
    int itry = 0;
    struct in_addr serverip;
    serverip.s_addr = 0;
    
    if( try != NULL ) {
        itry = *try;
    }
    
    printout( "Connecting to remote storage server...\n" );
    
    while( itry < RS_MAX_TRYSERVER ) {
        debug( "connecting to dht server, try #%d", itry );
        printout( "trying server %d\n", itry + 1 );
            
        itry++;

        if( ( serverip = find_ip( opts.server_name ) ).s_addr == 0 ) {
            continue;
        }
        
        if( opendht_connect( serverip, opts.server_port ) == DHT_SUCCESS ) {
            break;
        }
    }
    
    //if we're still disconnected, try the gateways list
    if( opendht_isconnected() == FALSE && gateways_isloaded() == TRUE ) {
        while( itry < RS_MAX_TRYSERVERCACHE ) {
            debug( "connecting to cached server, try #%d", itry );
            printout( "trying server %d\n", itry + 1 );
            
            itry++;
            
            if( ( serverip = find_from_gateways() ).s_addr == 0 ) {
                continue;
            }
            
            if( opendht_connect( serverip , opts.server_port ) == DHT_SUCCESS ) {
                break;
            }
        }
    }

    //update the caller's counter
    if( try != NULL ) {
        *try = itry;
    }
    
    if( opendht_isconnected() == TRUE ) {
        debug( "connected to server %s", inet_ntoa( serverip ) );
        printout( "Succesfully connected to remote storage server\n" ); 
        return SUCCESS;
    }
    else {
        debug( "failed to connect to a remote storage server" );
        printout( "Failed to connect to remote storage server\n" ); 
        return FAILURE;
    }
}
#endif

/*
 * If the gateways file was loaded, we use it.
 *
 * On the first try, we use the server on top of the list. If that fails, we
 * pick a random one on subsequent tries.
 *
 */ 
static int try_connect( int* try )
{
    int itry = ( try == NULL ) ? 0:(*try);
    struct in_addr serverip;
    memset( &serverip, 0x00, sizeof( struct in_addr ) );

    printout( "Connecting to remote storage server...\n" );
    
    // if the gateways list resources was loaded, we use it. otherwise, we rely
    // on opts.servername
    if( gateways_isloaded() == TRUE ) {
        while( opendht_isconnected() == FALSE && itry < RS_MAX_TRYSERVERCACHE ) {
            debug( "connecting to cached server, try #%d", itry );
            printout( "Trying server %d...please be patient\n", itry + 1 );
            
            serverip = find_from_gateways( itry );
            itry++;
            
            if( serverip.s_addr == 0 ) {
                continue;
            }

            if( opendht_connect( serverip , opts.server_port ) == DHT_SUCCESS ) {
                break;
            }
        }
    }
    else {
        while( opendht_isconnected() == FALSE && itry < RS_MAX_TRYSERVER ) {
            debug( "connecting to dht server, try #%d", itry );
            printout( "Trying server %d...please be patient.\n", itry + 1 );
                
            if( STROPT_ISLOADED( opts.server_name ) ) {
                serverip = find_ip( opts.server_name );
            }
            itry++;
            
            if( serverip.s_addr == 0 ) {
                continue;
            }
            
            if( opendht_connect( serverip, opts.server_port ) == DHT_SUCCESS ) {
                break;
            }
        }
    }
    
    //update the caller's counter
    if( try != NULL ) {
        *try = itry;
    }

    if( opendht_isconnected() == TRUE ) {
        debug( "connected to server %s", inet_ntoa( serverip ) );
        printout( "Succesfully connected to remote storage server\n" ); 
        return SUCCESS;
    }
    else {
        debug( "failed to connect to a remote storage server" );
        printout( "Could not connect to remote storage server\n" ); 
        return FAILURE;
    }
}

static int try_put( dhtkey_t key, uchar* value, int len )
{
    assert( key != NULL && value != NULL && len > 0 && len <= OPENDHT_MAX_PUTLEN );
    
    int ret = 0;
    int ptry = 0;
    int ctry = 0;
    while( ptry++ < RS_MAX_PUTTRY
           && ( ret = opendht_put( key, value, len ) ) != DHT_SUCCESS )
    {
        opendht_disconnect();
        //if we can't reconnect, we give up
        if( try_connect( &ctry ) == FAILURE ) {
            error( "could not reconnect" );
            break;
        }
    }
    
    return ( ret == DHT_SUCCESS ? SUCCESS:FAILURE );    
}

/*
 * Note: Please deallocate the returned value
 *
 */ 
static chunktrak_t* reassemble_chunks( getiter_t* iter, int* rtotalchunks, int* rtotalchunksfound, int* rtotalbytesfound )
{
    assert( rtotalchunks != NULL && rtotalchunksfound != NULL && rtotalbytesfound != NULL );

    chunktrak_t* tracker = NULL; // we place ptrs to the chunks here
    int totalbytesfound = 0;
    int totalchunksfound = -1;
    int totalchunks = 0;
    int firstheader = FALSE;
    
    while( iter->curroff < iter->len  && totalchunksfound < totalchunks ) {
        
        //seek to a header
        chunkhdr_t header;
        memset( &header, 0x00, sizeof( chunkhdr_t ) );

        while( iter->curroff + sizeof( chunkhdr_t ) < iter->len ) {
            deserialize_header( &header, iter->data + iter->curroff );
            if( memcmp( header.ident, RS_CHUNKHDR_MAGIC, RS_CHUNKHDR_MAGIC_LEN ) == 0 ) {
                iter->curroff += sizeof( chunkhdr_t );
                break;
            }
            
            iter->curroff++;
        }

        if( firstheader == FALSE ) {
            if( memcmp( header.ident, RS_CHUNKHDR_MAGIC, RS_CHUNKHDR_MAGIC_LEN ) != 0 ) {
                warn( "could not find a header" );
                break;
            }
            else if( memcmp( header.ident, RS_CHUNKHDR_MAGIC, RS_CHUNKHDR_MAGIC_LEN ) == 0 ) {
                int trackersize = header.numchunks * sizeof( chunktrak_t ); 
                tracker = (chunktrak_t*)malloc( trackersize  );
                assert( tracker != NULL );
                memset( tracker, 0x00, trackersize );
                totalchunks = header.numchunks; // correct if all data under this key is from this header
                totalchunksfound = 0;
                firstheader = TRUE;
            }
        }
        
        //now track this chunk
        if( memcmp( header.ident, RS_CHUNKHDR_MAGIC, RS_CHUNKHDR_MAGIC_LEN ) == 0 ) {
            tracker[header.currchunk].data = iter->data + iter->curroff;
            tracker[header.currchunk].dlen = header.chunklen;
            totalbytesfound += header.chunklen;
            iter->curroff += header.chunklen;
            totalchunksfound++;
        }
    }
    
    *rtotalchunks = totalchunks;
    *rtotalchunksfound = totalchunksfound;
    *rtotalbytesfound = totalbytesfound;
    return tracker;
}

int remstore_connect( void )
{
    if( opts.server_port != OPENDHT_PORT ) {
        warn( "option port %d is not SunRPC port for OpenDHT", opts.server_port );
    }
    
    //update gateways file
    update_gateways_list();

    return try_connect( NULL );
}

int remstore_isconnected( void )
{
    return opendht_isconnected();
}

int remstore_disconnect( void )
{
    if( opendht_disconnect() == DHT_SUCCESS && opendht_isconnected() == FALSE ) {
        debug( "disconnected from remote storage server" );
        return SUCCESS;
    }
    else {
        debug( "failed to disconnect from remote storage server" );
        return FAILURE;
    }
}

/*
 *
 * Insert <data> to DHT under <key>.
 * Splits <data> into chunks and appends them with a header to ensure
 * proper retrieval.
 *
 */ 
int remstore_put( dhtkey_t key, uchar* data IN, int dlen IN )
{
    if( opendht_isconnected() == FALSE ) {
        error( "not connected" );
        return FAILURE;
    }

    if( key != NULL && data != NULL && dlen > 0 ) {
        int i;
        int nfullchunks = dlen / RS_MAX_PUTLEN;
        int lastchunklen = dlen % RS_MAX_PUTLEN;

        uchar chunk[OPENDHT_MAX_PUTLEN]; //header + data
        uchar* curdata = data;
        chunkhdr_t header = { RS_CHUNKHDR_MAGIC, 0, 0, 0 };
        header.numchunks = ( lastchunklen == 0 ) ? nfullchunks:( nfullchunks + 1 );
        
        debug( "splitting %d into %d chunks", dlen, header.numchunks );
        
        for( i=0; i<nfullchunks; i++ ) {
            header.currchunk = i;
            header.chunklen = RS_MAX_PUTLEN; 
            serialize_header( &header, chunk );
            memcpy( &chunk[sizeof( chunkhdr_t )], curdata, header.chunklen );
            curdata += header.chunklen;
            
            //if we failed to put this chunk, then this key is a failure
            int totalchunklen = sizeof( chunkhdr_t ) + header.chunklen;
            debug( "chunk %d insert trying", header.currchunk );
            if( try_put( key, chunk, totalchunklen ) == FAILURE ) {
                debug( "chunk %d insert failed", header.currchunk );
                return FAILURE;
            }
            debug( "chunk %d insert successful", header.currchunk );
        }
        
        //if we got here, we are successful up to the second to the
        //last chunk. we try to insert the last chunk but if it fails,
        //we should still consider this key a failure
        if( lastchunklen != 0 ) {
            header.currchunk = header.numchunks - 1;
            header.chunklen = lastchunklen;
            serialize_header( &header, chunk );
            memcpy( &chunk[sizeof( chunkhdr_t )], curdata, header.chunklen );
            
            int totalchunklen = sizeof( chunkhdr_t ) + header.chunklen;
            debug( "last chunk %d insert trying", header.currchunk );
            if( try_put( key, chunk, totalchunklen ) == FAILURE ) {
                debug( "last chunk %d insert failed", header.currchunk );
                return FAILURE;
            }
            debug( "last chunk %d insert successful", header.currchunk );
        }
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/* 
 * Tries to get all the data under <key>
 * Callee should be sure to call remstore_getiter_done once finished
 *
 */
int remstore_getiter_init( uchar* key IN, getiter_t* iter OUT )
{
    if( key != NULL && iter != NULL ) {
        
        iter->data = NULL;
        iter->len = 0;
        
        int ret = 0;
        int gtry = 0;
        int ctry = 0;
        while( gtry++ < RS_MAX_GETTRY
               && ( ret = opendht_get( key, &iter->data, &iter->len ) ) != DHT_SUCCESS )
        {
            opendht_disconnect();
            if( ret == DHT_TIMEOUT ) {
                printout( "Connection timed out...retrying\n" );
            }
            else {
                printout( "Connection ended...retrying\n" );
            }
            
            if( try_connect( &ctry ) == FAILURE ) {
                error( "could not reconnect" );
                break; //fall into DHT_SUCCESS check below
            }
        }

        //incomplete data, we don't want this? not sure...
        if( ret != DHT_SUCCESS ) {
            if( iter->data != NULL ) {
                memset( iter->data, 0x00, iter->len );
                free( iter->data );
                iter->data = NULL;
            }
            
            iter->len = 0;
        
            return FAILURE;
        }
        
        //succesfully got all data under key
        iter->curroff = 0;
        
        return SUCCESS;
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

/*
 *
 * Reassembles chunks. They are assumed to be out of order.
 *
 * Note: It could be the case that we inserted twice to a key and
 * this routine will not handle that case correctly i.e. all ADEONA
 * headers are the same. The probability is small so we don't care. 
 *
 */
int remstore_getiter_next_alloc( getiter_t* iter IN, uchar** data OUT, int* dlen OUT )
{
    if( data != NULL && iter != NULL && *data == NULL ) {
        
        int totalbytesfound = 0;
        int totalchunksfound = -1;
        int totalchunks = 0;
        
        chunktrak_t* tracker = reassemble_chunks( iter, &totalchunks, &totalchunksfound, &totalbytesfound ); 
        debug( "total chunks: %d, total chunks found: %d", totalchunks, totalchunksfound );
        debug( "total bytes found: %d", totalbytesfound );
        
        if( tracker == NULL ) {
            return FAILURE;
        }
        
        //write the chunks into <data> only if we found all the
        //chunks, otherwise this retrieve is a failure
        if( totalchunksfound == totalchunks ) {
            *data = (uchar*)malloc( totalbytesfound );
            assert( *data != NULL );

            uchar* dptr = *data;
            int i;
            for( i=0; i<totalchunks; i++ ) {
                if( tracker[i].data == NULL ) { 
                    error( "missing chunk %d", i );
                    free( *data );
                    *data = NULL;
                    *dlen = 0;
                    free( tracker );
                    return FAILURE;
                }

                memcpy( dptr, tracker[i].data, tracker[i].dlen );
                dptr += tracker[i].dlen;
            }
            
            assert( ( dptr - *data ) == totalbytesfound );
            *dlen = totalbytesfound;
            
            free( tracker );
            return SUCCESS;
        }
        else {
            //no data or incomplete data
            return FAILURE;
        }
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }

}

int remstore_getiter_next( getiter_t* iter IN, uchar* data OUT, int dlen IN )
{
    if( iter != NULL && data != NULL && dlen > 0 ) {
        int totalbytesfound = 0;
        int totalchunksfound = -1;
        int totalchunks = 0;
        
        chunktrak_t* tracker = reassemble_chunks( iter, &totalchunks, &totalchunksfound, &totalbytesfound ); 
        debug( "total chunks: %d, total chunks found: %d", totalchunks, totalchunksfound );
        debug( "total bytes found: %d", totalbytesfound );
        
        if( tracker == NULL ) {
            return FAILURE;
        }

        //write the chunks into <data> only if we found all the
        //chunks, otherwise this retrieve is a failure
        if( totalchunksfound == totalchunks ) {
            if( dlen < totalbytesfound ) {
                warn( "space for results not enough, dlen = %d < totalbytesfound = %d", dlen, totalbytesfound );
            }
            
            uchar* dptr = data;
            int i;
            for( i=0; i<totalchunks; i++ ) {
                if( tracker[i].data == NULL ) { 
                    error( "missing chunk %d", i );
                    free( tracker );
                    return FAILURE;
                }
                
                if( ( dptr + tracker[i].dlen ) - data > dlen ) {
                    warn( "not enough space for chunk %d, ignoring rest of chunks", i );
                    i = -1; //for debug
                }

                memcpy( dptr, tracker[i].data, tracker[i].dlen );
                dptr += tracker[i].dlen;
            }
            
            if( i > 0 ) {
                assert( ( dptr - data ) == totalbytesfound );
            }

            free( tracker );
            return SUCCESS;
        }
        else {
            //no data or incomplete data
            return FAILURE;
        }
        
    }
    else {
        error( "invalid input" );
        return FAILURE;
    }
}

int remstore_getiter_done( getiter_t* iter IN )
{
    if( iter != NULL && iter->data != NULL ) {
        memset( iter->data, 0x00, iter->len );
        free( iter->data );
        iter->data = NULL;
        memset( iter, 0x00, sizeof( getiter_t ) );
    }
    
    return SUCCESS;
}

